"""Verbs ported from r-stats"""
from typing import Iterable

from pandas import DataFrame, isna
from pipda import register_verb

from ..core.defaults import NA_REPR
from ..core.grouped import DataFrameGroupBy


@register_verb(DataFrame)
def set_names(obj: DataFrame, names: Iterable[str]) -> DataFrame:
    """Set names of a dataframe"""
    names = [NA_REPR if isna(name) else name for name in names]
    obj = obj.copy()
    obj.columns = names
    return obj


@set_names.register(DataFrameGroupBy)
def _(obj: DataFrameGroupBy, names: Iterable[str]) -> DataFrameGroupBy:
    """Set names of a grouped/rowwise df"""
    obj = obj.copy(copy_grouped=True)
    names_dict = dict(zip(obj.columns, names))
    obj.columns = names
    gvars = [names_dict[name] for name in obj.attrs['_group_vars']]
    obj.attrs["_group_vars"] = gvars
    obj.attrs["_group_data"] = obj._group_data.copy()
    obj._group_data.columns = gvars + [obj._group_data.columns[-1]]
    return obj



"""Load operator, provide f and __version__"""
from typing import Mapping

from .core import operator as _
from .core import _frame_format_patch
from .core.defaults import f

__all__ = ('f', 'get_versions')
__version__ = "0.5.4"


def get_versions(
    prnt: bool = True
) -> Mapping[str, str]:  # pragma: no cover
    """Print or return related versions which help for bug reporting.

    Args:
        prnt: Print the versions instead of returning them?

    Returns:
        A `Diot` object of the versions.
    """
    import sys

    import numpy
    import pandas
    import pipda
    import executing
    import varname
    from diot import Diot

    out = Diot(
        python=sys.version,
        datar=__version__,
        numpy=numpy.__version__,
        pandas=pandas.__version__,
        pipda=pipda.__version__,
        executing=executing.__version__,
        varname=varname.__version__,
    )
    if not prnt:
        return out

    keylen = max(map(len, out))
    for key, ver in out.items():
        verlines = ver.splitlines()
        print(f"{key.ljust(keylen)}: {verlines.pop(0)}")
        for verline in verlines:
            print(f"{' ' * keylen}  {verline}")

    return None



"""APIs ported from R-tidyr"""

from .funcs import full_seq
from .chop import chop, unchop
from .nest import nest, unnest
from .pack import pack, unpack
from .expand import expand_grid, expand, nesting, crossing
from .complete import complete
from .drop_na import drop_na
from .extract import extract
from .fill import fill
from .pivot_long import pivot_longer
from .pivot_wide import pivot_wider
from .separate import separate, separate_rows
from .uncount import uncount
from .unite import unite
from .replace_na import replace_na



"""Provide Collection and related classes to mimic `c` from `r-base`"""

from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Union

import pandas
from pipda import evaluate_expr
from pipda.context import ContextAnnoType

from .utils import get_option
from .types import is_iterable, is_scalar, is_scalar_int
from .exceptions import ColumnNotExistingError

PoolType = Union[Iterable, int]
UNMATCHED = object()


class CollectionBase(ABC):
    """Abstract class for collections"""

    def __init__(
        self, *args: Any, pool: PoolType = None, base0: bool = None
    ) -> None:
        self.elems = args
        self.base0 = base0
        self.pool = pool
        self.unmatched = set()
        self.error = None
        try:
            self.expand(pool=pool, base0=self.base0)
        except (ValueError, ColumnNotExistingError) as exc:
            self.error = exc

    def _pipda_eval(self, data: Any, context: ContextAnnoType) -> Any:
        """Defines how the object should be evaluated when evaluated by
        pipda's evaluation"""
        self.elems = evaluate_expr(self.elems, data, context)
        return self

    @abstractmethod
    def expand(
        self, pool: PoolType = None, base0: bool = None
    ) -> "CollectionBase":
        """Expand the collection"""


class Collection(CollectionBase, list):
    """Mimic the c function in R

    All elements will be flattened.
    This is also the entry point to adopt 1-baed and 0-based indexing, and
    convert them into 0-based finally

    The Inverted, Negated and slice objects will be expanded immediately. This
    means there is no chance to apply `base0_` that is received later on. So
    the original elements are stored in `self.elems` to wait for a second
    evaluation with the correct `base0_`.

    Args:
        *args: The elements
        base0: Whether the index is 0-based or not. Should be an integer
            indicating the range or a list, not a generator.
        pool: The pool used to expand slice
    """
    def _get_base0(self, base0: bool) -> bool:
        """Get base0 if specified, otherwise self.base0"""
        if base0 is None:
            base0 = self.base0
        self.base0 = get_option("index.base.0", base0)
        return self.base0

    def _get_pool(self, pool: PoolType) -> PoolType:
        """Get pool if specified, otherwise self.pool"""
        if pool is not None:
            self.pool = pool
        return self.pool

    def expand(
        self, pool: PoolType = None, base0: bool = None
    ) -> CollectionBase:
        """Expand the elements of this collection

        A element could be either
        - a string (a column name),
        - an integer (a literal number, or an index of the pool)
        - a Collection
        - a Negated object
        - an Inverted object
        - a slice object

        If `base0` is passed, then it is used, otherwise use `self.base0`
        """
        base0 = self._get_base0(base0)
        pool = self._get_pool(pool)
        self.unmatched.clear()
        self.error = None

        if pool is not None:
            elems = [
                elem
                for elem in self.elems
                if not is_scalar(elem) or not pandas.isnull(elem)
            ]
        else:
            elems = self.elems  # type: ignore

        if not elems:
            list.__init__(self, [])
            return self

        inverts = [isinstance(elem, Inverted) for elem in elems]
        if all(inverts):
            list.__init__(
                self,
                Inverted(
                    Collection(*(elem.elems for elem in elems)),
                    pool=pool,
                    base0=base0,
                ),
            )
            return self

        if any(inverts):
            raise ValueError(
                "Cannot mix Inverted and non-Inverted elements "
                "in a collection.."
            )

        expanded = []
        expanded_append = expanded.append
        expanded_extend = expanded.extend
        for elem in elems:
            if isinstance(elem, slice):
                elem = Slice(elem, pool=pool, base0=base0)

            if isinstance(elem, CollectionBase):
                expanded_extend(elem.expand(pool, base0))
                self.unmatched.update(elem.unmatched)
            elif is_scalar(elem):
                elem = self._index_from_pool(elem)
                if elem is not UNMATCHED:
                    expanded_append(elem)
            else:  # iterable
                exp = Collection(*elem, pool=pool, base0=base0)
                self.unmatched.update(exp.unmatched)
                expanded_extend(exp)
        list.__init__(self, expanded)
        return self

    def _is_index(self, elem: Any) -> bool:
        """Check if an element is an index or not"""
        if self.pool is None or not is_scalar_int(elem):
            return False
        if is_scalar_int(self.pool):
            return True
        # iterable
        # If an empty pool is given, assuming integer elem is index
        # Otherwise if pool is a list of integers, then elem should match
        # the pool
        if len(self.pool) == 0 or not is_scalar_int(self.pool[0]):
            return True
        return False

    def _index_from_pool(self, elem: Any, base0: bool = None) -> Any:
        """Try to pull the index of the element from the pool"""
        if self.pool is None:
            # Return the element itself if pool is not specified
            # Then element is supposed to be a literal
            return elem

        if base0 is None:
            base0 = self.base0

        if self._is_index(elem):
            # elem is treated as an index if it is not an element of the pool
            if not base0 and elem == 0:
                raise ValueError("Index 0 given for 1-based indexing.")

            pool = len(self.pool) if is_iterable(self.pool) else self.pool
            out = elem - int(not base0) if elem >= 0 else elem + pool
            # then the index should be 0 ~ len-1
            if not 0 <= out < pool:
                self.unmatched.add(elem)
                return UNMATCHED

            return out

        pool = range(self.pool) if is_scalar(self.pool) else self.pool

        if elem not in pool:
            self.unmatched.add(elem)
            return UNMATCHED

        return pool.index(elem)

    def __repr__(self) -> str:
        return f"Collection({self.elems})"

    def __str__(self) -> str:
        return list.__repr__(self)

    def __neg__(self):
        return Negated(self, pool=self.pool, base0=self.base0)

    def __invert__(self):
        return Inverted(self, pool=self.pool, base0=self.base0)


class Negated(Collection):
    """Negated collection, representing collections by `-c(...)` or `-f[...]`"""

    def __repr__(self) -> str:
        return f"Negated({self.elems})"

    def expand(self, pool: PoolType = None, base0: bool = None) -> None:
        """Expand the object"""
        super().expand(pool, base0)
        # self is now 0-based indexes

        if pool is not None:
            elems = [
                self._index_from_pool(-elem - int(not base0), base0=True)
                for elem in reversed(self)
            ]
            # for elem in reversed(self):
            #     # If matched, it's sure an index.
            #     # Unmatched elements are stored in self.unmatched
            #     # if not self._is_index(elem):
            #     #     raise ValueError("Cannot negate a non-index value.")
            #     elems.append(-elem-1)
            list.__init__(self, elems)
        else:
            list.__init__(self, [-elem for elem in self])
        return self


class Inverted(Collection):
    """Inverted collection, tries to exlude some elements"""

    def __repr__(self) -> str:
        return f"Inverted({self.elems})"

    def expand(self, pool: PoolType = None, base0: bool = None) -> None:
        """Expand the object"""
        if pool is None:
            raise ValueError("Inverted object needs `pool` to expand.")

        super().expand(pool, base0)  # 0-based indexes
        pool = range(pool) if is_scalar_int(pool) else range(len(pool))

        list.__init__(self, [elem for elem in pool if elem not in self])
        return self


class Intersect(Collection):
    """Intersect of two collections, designed for `&` operator"""

    def __init__(
        self, *args: Any, pool: PoolType = None, base0: bool = None
    ) -> None:
        if len(args) != 2:
            raise ValueError("Intersect can only accept two collections.")
        self.elems = args
        self.base0 = base0
        self.pool = pool
        self.unmatched = set()
        self.error = None
        # don't expand.

    def __repr__(self) -> str:
        return f"Intersect({self.elems})"

    def expand(self, pool: PoolType = None, base0: bool = None) -> None:
        """Expand the object"""
        left = Collection(self.elems[0], pool=pool, base0=base0)
        right = Collection(self.elems[1], pool=pool, base0=base0)
        list.__init__(self, [elem for elem in left if elem in right])
        return self


class Slice(Collection):
    """Slice to wrap builtins.slice

    to sanitize slice objects, and compile it into a list of indexes

    If pool if given, then slice object will try to match the elements
    in pool.
    """

    def __init__(
        self, *args: Any, pool: PoolType = None, base0: bool = None
    ) -> None:
        if len(args) != 1 or not isinstance(args[0], slice):
            raise ValueError("Slice should wrap one and only one slice object.")
        self.slc = args[0]
        super().__init__(*args, pool=pool, base0=base0)

    def __repr__(self) -> str:
        return f"Slice({self.elems})"

    def expand(self, pool: PoolType = None, base0: bool = None) -> None:
        base0 = self._get_base0(base0)
        pool = self._get_pool(pool)
        self.unmatched.clear()
        self.error = None

        if pool is None:
            expanded = self._expand_no_pool(base0)
        else:
            expanded = self._expand_pool(pool, base0)

        list.__init__(self, expanded)
        return self

    def _expand_no_pool(self, base0: bool) -> List[Any]:
        """Expand slice literally

        Without pool or length, `[:3]` will expand to `0,1,2`,
        and `[-3:]` `-3,-2,-1`. But with length 10, `[-3:]` will expand to
        `7,8,9`
        """
        start, stop, step = self.slc.start, self.slc.stop, self.slc.step
        # Without pool, we don't know how to interpret strings in slice
        if isinstance(start, str) or isinstance(stop, str):
            raise ValueError(
                "`pool` is required when start/stop of slice are not indexes."
            )

        if start is None:
            start = int(not base0)
        if stop is None:
            stop = int(not base0)
        if step is None:
            step = 1 if stop >= start else -1

        out = []
        out_append = out.append
        i = start
        while (i < stop) if step > 0 else (i > stop):
            out_append(i)
            i += step
        if not base0:  # include stop
            out_append(stop)
        return out

    def _expand_pool(self, pool: PoolType, base0: bool) -> List[Any]:
        """Slice with pool given; try to match the range with the elements
        in the pool"""
        start, stop, step = self.slc.start, self.slc.stop, self.slc.step
        base = int(not base0)

        if start is None:
            start = 0
        elif not is_scalar_int(start):
            if is_scalar_int(pool) or start not in pool:
                raise ColumnNotExistingError(
                    f"Column `{start}` does not exist."
                )
            start = self._index_from_pool(start)
        else:
            start -= base

        len_pool = pool if is_scalar_int(pool) else len(pool)
        if stop is None:
            stop = len_pool
        elif not is_scalar_int(stop):
            if is_scalar_int(pool) or stop not in pool:
                raise ColumnNotExistingError(f"Column `{stop}` does not exist.")
            stop = self._index_from_pool(stop) + 1
        # else:
        #     stop += base

        if step == 0:
            stop -= 1
            step = 1 if stop >= start else -1

        return list(range(*slice(start, stop, step).indices(len_pool)))



"""Public APIs from this submodule"""
from ..core.options import get_option, options, options_context
from .arithmetic import (
    ceiling,
    cov,
    floor,
    mean,
    median,
    pmax,
    pmin,
    sqrt,
    var,
    scale,
    col_sums,
    row_sums,
    col_means,
    row_means,
    col_sds,
    row_sds,
    col_medians,
    row_medians,
    min as min_,
    max as max_,
    round as round_,
    sum as sum_,
    abs as abs_,
    prod,
    sign,
    signif,
    trunc,
    exp,
    log,
    log2,
    log10,
    log1p,
)
from .bessel import bessel_i, bessel_j, bessel_k, bessel_y
from .casting import as_double, as_float, as_int, as_integer, as_numeric
from .complex import arg, as_complex, conj, im, is_complex, mod, re as re_
from .constants import LETTERS, letters, month_abb, month_name, pi
from .cum import cummax, cummin, cumprod, cumsum
from .date import as_date, as_pd_date
from .factor import (
    as_categorical,
    as_factor,
    droplevels,
    factor,
    is_categorical,
    is_factor,
    is_ordered,
    levels,
    nlevels,
    ordered,
)
from .funs import (
    cut,
    data_context,
    expandgrid,
    identity,
    make_unique,
    make_names,
    rank,
)
from .logical import (
    FALSE,
    TRUE,
    as_bool,
    as_logical,
    is_bool,
    is_false,
    is_logical,
    is_true,
)
from .na import NA, NaN, any_na, is_na, Inf, is_finite, is_infinite, is_nan
from .null import NULL, as_null, is_null
from .random import set_seed
from .seq import (
    c,
    length,
    lengths,
    order,
    rep,
    rev,
    sample,
    seq,
    seq_along,
    seq_len,
    sort,
    unique,
    match,
)
from .special import (
    beta,
    choose,
    digamma,
    factorial,
    gamma,
    lbeta,
    lchoose,
    lfactorial,
    lgamma,
    psigamma,
    trigamma,
)
from .string import (
    as_character,
    as_str,
    as_string,
    grep,
    grepl,
    gsub,
    is_character,
    is_str,
    is_string,
    nchar,
    nzchar,
    paste,
    paste0,
    sprintf,
    strsplit,
    sub,
    substr,
    substring,
    startswith,
    endswith,
    strtoi,
    trimws,
    chartr,
    tolower,
    toupper,
)
from .table import table, tabulate
from .testing import (
    is_atomic,
    is_double,
    is_element,
    is_float,
    is_in,
    is_int,
    is_integer,
    is_numeric,
    any as any_,
    all as all_,
)
from .trig_hb import (
    acos,
    acosh,
    asin,
    asinh,
    atan,
    atan2,
    atanh,
    cos,
    cosh,
    cospi,
    sin,
    sinh,
    sinpi,
    tan,
    tanh,
    tanpi,
)
from .verbs import (
    append,
    colnames,
    diag,
    dim,
    duplicated,
    intersect,
    names,
    ncol,
    nrow,
    prop_table,
    proportions,
    rownames,
    setdiff,
    setequal,
    t,
    union,
    max_col,
    complete_cases,
)
from .which import which, which_max, which_min

__all__ = [name for name in locals() if not name.startswith("_")]

_builtin_names = {
    "min": min_,
    "max": max_,
    "sum": sum_,
    "abs": abs_,
    "round": round_,
    "all": all_,
    "any": any_,
    "re": re_,
}
__all__.extend(_builtin_names)

# warn when builtin names are imported directly
from ..core.warn_builtin_names import warn_builtin_names

__getattr__ = warn_builtin_names(**_builtin_names)



"""Group by verbs and functions
See source https://github.com/tidyverse/dplyr/blob/master/R/group-by.r
"""

from typing import Any, Mapping, Union

from pandas import DataFrame
from pipda import register_verb
from pipda.symbolic import DirectRefAttr, DirectRefItem
from pipda.expression import Expression
from pipda.utils import CallingEnvs

from ..core.grouped import DataFrameGroupBy, DataFrameRowwise
from ..core.contexts import Context
from ..core.defaults import f
from ..core.utils import (
    copy_attrs,
    name_mutatable_args,
    check_column_uniqueness,
    reconstruct_tibble,
    vars_select,
)
from ..core.exceptions import ColumnNotExistingError
from ..base import setdiff, union

from .group_data import group_vars


@register_verb(DataFrame, context=Context.PENDING)
def group_by(
    _data: DataFrame,
    *cols: Any,
    _add: bool = False,  # not working, since _data is not grouped
    _drop: bool = None,
    **kwargs: Any,
) -> DataFrameGroupBy:
    """Takes an existing tbl and converts it into a grouped tbl where
    operations are performed "by group"

    See https://dplyr.tidyverse.org/reference/group_by.html

    Note that this does not return `pandas.DataFrameGroupBy` object but a
    `datar.core.grouped.DataFrameGroupBy` object, which is a subclass of
    `DataFrame`. This way, it will be easier to implement the APIs that related
    to grouped data.

    Args:
        _data: The dataframe
        _add: When False, the default, `group_by()` will override
            existing groups. To add to the existing groups, use `_add=True`.
        _drop: Drop groups formed by factor levels that don't appear in the
            data? The default is True except when `_data` has been previously
            grouped with `_drop=False`.
        *cols: variables or computations to group by.
            Note that columns here cannot be selected by indexes. As they are
            treated as computations to be added as new columns.
            So no `base0_` argument is supported.
        **kwargs: Extra variables to group the dataframe

    Return:
        A `datar.core.grouped.DataFrameGroupBy` object
    """
    if not cols and not kwargs:
        return reconstruct_tibble(_data, _data)

    if _drop is None:
        _drop = group_by_drop_default(_data)

    groups = _group_by_prepare(_data, *cols, **kwargs, _add=_add)
    out = DataFrameGroupBy(
        groups["data"],
        _group_vars=groups["group_names"],
        _group_drop=_drop,
    )
    copy_attrs(out, _data)
    return out


@register_verb(DataFrame, context=Context.SELECT)
def rowwise(
    _data: DataFrame, *cols: Union[str, int], base0_: bool = None
) -> DataFrameRowwise:
    """Compute on a data frame a row-at-a-time

    See https://dplyr.tidyverse.org/reference/rowwise.html

    Args:
        _data: The dataframe
        *cols:  Variables to be preserved when calling summarise().
            This is typically a set of variables whose combination
            uniquely identify each row.
        base0_: Whether indexes are 0-based if columns are selected by indexes.
            If not given, will use `datar.base.get_option('index.base.0')`

    Returns:
        A row-wise data frame
    """
    check_column_uniqueness(_data)
    idxes = vars_select(_data.columns, *cols, base0=base0_)
    if len(idxes) == 0:
        return DataFrameRowwise(_data)
    return DataFrameRowwise(_data, _group_vars=_data.columns[idxes].tolist())


@rowwise.register(DataFrameGroupBy, context=Context.SELECT)
def _(
    _data: DataFrameGroupBy, *cols: str, base0_: bool = None
) -> DataFrameRowwise:
    # grouped dataframe's columns are unique already
    if cols:
        raise ValueError(
            "Can't re-group when creating rowwise data. "
            "Either first `ungroup()` or call `rowwise()` without arguments."
        )
    # copy_attrs?
    return DataFrameRowwise(
        _data,
        _group_vars=group_vars(_data, __calling_env=CallingEnvs.REGULAR),
    )


@rowwise.register(DataFrameRowwise, context=Context.SELECT)
def _(
    _data: DataFrameRowwise, *cols: str, base0_: bool = None
) -> DataFrameRowwise:
    idxes = vars_select(_data.columns, *cols, base0=base0_)
    if len(idxes) == 0:
        # copy_attrs?
        return DataFrameRowwise(_data)
    # copy_attrs?
    return DataFrameRowwise(_data, _group_vars=_data.columns[idxes].to_list())


@register_verb(DataFrame, context=Context.SELECT)
def ungroup(
    x: DataFrame, *cols: Union[str, int], base0_: bool = None
) -> DataFrame:
    """Ungroup a grouped data

    See https://dplyr.tidyverse.org/reference/group_by.html

    Args:
        x: The data frame
        *cols: Variables to remove from the grouping variables.
        base0_: If columns are selected with indexes, whether they are 0-based.
            If not given, will use `datar.base.get_option('index.base.0')`

    Returns:
        A data frame with selected columns removed from the grouping variables.
    """
    if cols:
        raise ValueError("`*cols` is not empty.")
    return x


@ungroup.register(DataFrameGroupBy, context=Context.SELECT)
def _(
    x: DataFrameGroupBy, *cols: Union[str, int], base0_: bool = None
) -> DataFrame:
    if not cols:
        return DataFrame(x, index=x.index)
    old_groups = group_vars(x, __calling_env=CallingEnvs.REGULAR)
    to_remove = vars_select(x.columns, *cols, base0=base0_)
    new_groups = setdiff(
        old_groups, x.columns[to_remove], __calling_env=CallingEnvs.REGULAR
    )

    return group_by(x, *new_groups, __calling_env=CallingEnvs.REGULAR)


@ungroup.register(DataFrameRowwise, context=Context.SELECT)
def _(
    x: DataFrameRowwise, *cols: Union[str, int], base0_: bool = None
) -> DataFrame:
    if cols:
        raise ValueError("`*cols` is not empty.")
    return DataFrame(x)


def _group_by_prepare(
    _data: DataFrame, *args: Any, _add: bool = False, **kwargs: Any
) -> Mapping[str, Any]:
    """Prepare for group by"""
    computed_columns = _add_computed_columns(
        _data, *args, _fn="group_by", **kwargs
    )

    out = computed_columns["data"]
    group_names = computed_columns["added_names"]
    if _add:
        group_names = union(
            group_vars(_data, __calling_env=CallingEnvs.REGULAR),
            group_names,
            __calling_env=CallingEnvs.REGULAR,
        )

    # checked in _add_computed_columns
    # unknown = setdiff(group_names, out.columns)
    # if unknown:
    #     raise ValueError(
    #         "Must group by variables found in `_data`.",
    #         f"Column `{unknown}` not found."
    #     )

    return {"data": out, "group_names": group_names}


def _add_computed_columns(
    _data: DataFrame, *args: Any, _fn: str = "group_by", **kwargs: Any
) -> Mapping[str, Any]:
    """Add mutated columns if necessary"""
    from .mutate import _mutate_cols

    # support direct strings
    args = [f[arg] if isinstance(arg, str) else arg for arg in args]
    named = name_mutatable_args(*args, **kwargs)
    if any(
        isinstance(val, Expression)
        and not isinstance(val, (DirectRefAttr, DirectRefItem))
        for val in named.values()
    ):
        context = Context.EVAL.value
        try:
            cols, nonexists = _mutate_cols(
                ungroup(_data, __calling_env=CallingEnvs.REGULAR),
                context,
                *args,
                **kwargs,
            )
        except Exception as exc:
            raise ValueError(
                f"Problem adding computed columns in `{_fn}()`."
            ) from exc

        out = _data.copy()
        col_names = cols.columns.tolist()
        out[col_names] = cols
    else:
        out = _data
        col_names = list(named)
        nonexists = setdiff(
            col_names, out.columns, __calling_env=CallingEnvs.REGULAR
        )

    if nonexists:
        raise ColumnNotExistingError(
            "Must group by variables found in `_data`. "
            f"Columns {nonexists} are not found."
        )

    return {"data": out, "added_names": col_names}


def group_by_drop_default(_tbl) -> bool:
    """Get the groupby _drop attribute of dataframe"""
    return _tbl.attrs.get("_group_drop", True)



"""Port `table` function from r-base"""

from typing import Any, Iterable, Mapping, Tuple, Union, List

import numpy
import pandas
from pandas import Categorical, DataFrame, Series
from pipda import register_func
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.types import (
    ArrayLikeType,
    is_iterable,
    is_scalar,
    is_categorical,
    is_null,
)
from ..core.utils import Array, categorized, fillna_safe

from .na import NA


@register_func(None, context=Context.EVAL)
def table(
    input: Union[ArrayLikeType, Categorical, DataFrame],
    *more_inputs: Any,
    exclude: Any = NA,
    # use_na: str = "no", # TODO
    dnn: Union[str, List[str]] = None,
    # not supported, varname.argname not working with wrappers having
    # different signatures.
    # TODO: varname.argname2() now supports it
    # deparse_level: int = 1
) -> DataFrame:

    """uses the cross-classifying factors to build a contingency table of
    the counts at each combination of factor levels.

    When used with DataFrameGroupBy data, groups are ignored.

    Args:
        input: and
        *more_inputs: one or more objects which can be interpreted as factors
            Only 1 or 2 variables allowed currently.
            If obj or elements of objs is a DataFrame, each column is counted
            as a variable.
        exclude: levels to remove for all factors
            By default exclude NA. Set to None to not exclude anything.
        dnn: the names to be given to the dimensions in the result.

    Returns:
        A contingency table (DataFrame)
    """
    obj1, obj2 = _check_table_inputs(input, more_inputs)

    dn1 = dn2 = None
    if isinstance(dnn, str):
        dn1 = dn2 = dnn
    elif is_iterable(dnn):
        dnn = list(dnn)
        if len(dnn) == 1:
            dnn = dnn * 2
        dn1, dn2 = dnn[:2]

    if obj1 is obj2:
        obj1 = obj2 = _iterable_excludes(obj1, exclude=exclude)
    else:
        obj1 = _iterable_excludes(obj1, exclude=exclude)
        obj2 = _iterable_excludes(obj2, exclude=exclude)

    kwargs = {"dropna": False}  # type: Mapping[str, Any]
    if dn1:
        kwargs["rownames"] = [dn1]
    if dn2:
        kwargs["colnames"] = [dn2]

    tab = pandas.crosstab(obj1, obj2, **kwargs)
    if obj1 is obj2:
        tab = DataFrame(dict(count=numpy.diag(tab)), index=tab.columns).T
        tab.index.name = dn1
        tab.columns.name = dn2

    return tab


@register_func(None, context=Context.EVAL)
def tabulate(
    bin: Union[ArrayLikeType, Categorical],
    nbins: int = None,
) -> numpy.ndarray:
    """Takes the integer-valued vector `bin` and counts the
    number of times each integer occurs in it.

    Args:
        bin: A numeric vector (of positive integers), or a factor.
        nbins: the number of bins to be used.

    Returns:
        An integer valued ‘integer’ vector (without names).
        There is a bin for each of the values ‘1, ..., nbins’
    """
    from . import as_integer, t

    bin = as_integer(bin, base0_=False, __calling_env=CallingEnvs.REGULAR)
    nbins = max(1, max(bin) if len(bin) > 0 else 0, nbins)
    tabled = table(bin, __calling_env=CallingEnvs.REGULAR)
    tabled = (
        t(tabled, __calling_env=CallingEnvs.REGULAR)
        .reindex(range(1, nbins + 1), fill_value=0)
        .iloc[:, 0]
        .values
    )

    return tabled


def _check_table_inputs(
    input: Any, more_inputs: Any
) -> Tuple[Iterable, Iterable]:
    """Check and clean up `table` inputs"""
    too_many_input_vars_msg = "At most 2 iterables supported for `table`."
    obj1 = obj2 = None
    obj_nvar = 1

    if not isinstance(input, DataFrame):
        obj1 = list(input) if isinstance(input, str) else input
        obj1 = obj1 if is_categorical(obj1) else Array(obj1)
        obj_nvar = 1
    elif input.shape[1] == 0:
        raise ValueError("`input` of `table` has no columns.")
    elif input.shape[1] == 1:
        obj1 = input.iloc[:, 0]
        obj_nvar = 1
    elif input.shape[1] == 2:
        obj1 = input.iloc[:, 0]
        obj2 = input.iloc[:, 1]
        obj_nvar = 2
    else:
        raise ValueError(too_many_input_vars_msg)

    if obj_nvar == 2 and more_inputs:
        raise ValueError(too_many_input_vars_msg)

    if more_inputs:
        if len(more_inputs) > 1:
            raise ValueError(too_many_input_vars_msg)

        minput = more_inputs[0]
        if isinstance(minput, DataFrame):
            if minput.shape[1] > 1:
                raise ValueError(too_many_input_vars_msg)
            if minput.shape[1] == 0:
                raise ValueError("`*more_inputs` of `table` has no columns.")
            obj2 = minput.iloc[:, 0]
        elif isinstance(minput, str):
            obj2 = Array(list(minput))
        else:
            obj2 = minput if is_categorical(minput) else Array(minput)

    if obj2 is None:
        obj2 = obj1

    return obj1, obj2


def _iterable_excludes(data: Iterable, exclude: Iterable) -> Iterable:
    """Exclude values for categorical data"""
    if is_categorical(data) and exclude is NA:
        return data

    if exclude is None:
        return fillna_safe(data)

    if is_scalar(exclude):
        exclude = [exclude]

    exclude = Array(exclude)

    if is_categorical(data):
        data = categorized(data)
        data = data[~data.isin(exclude)]
        data = data.remove_categories(data.categories.intersection(exclude))

    else:
        data = Series(data)
        data = data[~data.isin(exclude)]

    return data if is_null(exclude).any() else fillna_safe(data)



"""Middlewares for datar"""
from typing import Any, Mapping, Tuple
from pipda.utils import DataEnv


class CurColumn:
    """Current column in across"""

    @classmethod
    def replace_args(cls, args: Tuple[Any], column: str) -> Tuple[Any, ...]:
        """Replace self with the real column in args"""
        return tuple(column if isinstance(arg, cls) else arg for arg in args)

    @classmethod
    def replace_kwargs(
        cls, kwargs: Mapping[str, Any], column: str
    ) -> Mapping[str, Any]:
        """Replace self with the real column in kwargs"""
        return {
            key: column if isinstance(val, cls) else val
            for key, val in kwargs.items()
        }


class WithDataEnv:
    """Implements `with data` to mimic R's `with(data, ...)`"""

    def __init__(self, data: Any) -> None:
        self.data = DataEnv(data)

    def __enter__(self) -> Any:
        return self.data

    def __exit__(self, *exc_info) -> None:
        self.data.delete()



"""Public APIs for dplyr"""

from .across import across, c_across, if_all, if_any
from .arrange import arrange
from .bind import bind_cols, bind_rows
from .context import (
    cur_column,
    cur_data,
    cur_data_all,
    cur_group,
    cur_group_id,
    cur_group_rows,
    n,
)
from .count_tally import add_count, add_tally, count, tally
from .desc import desc
from .dfilter import filter as filter_
from .distinct import distinct, n_distinct
from .dslice import slice as slice_
from .dslice import slice_head, slice_max, slice_min, slice_sample, slice_tail
from .funs import (
    between,
    coalesce,
    cumall,
    cumany,
    cummean,
    first,
    last,
    na_if,
    near,
    nth,
)
from .group_by import group_by, group_by_drop_default, rowwise, ungroup
from .group_data import (
    group_cols,
    group_data,
    group_indices,
    group_keys,
    group_rows,
    group_size,
    group_vars,
    n_groups,
)
from .group_iter import (
    group_map,
    group_modify,
    group_split,
    group_trim,
    group_walk,
    with_groups,
)
from .if_else import case_when, if_else
from .join import (
    anti_join,
    full_join,
    inner_join,
    left_join,
    nest_join,
    right_join,
    semi_join,
)
from .lead_lag import lag, lead
from .mutate import mutate, transmute
from .order_by import order_by, with_order
from .pull import pull
from .rank import (
    cume_dist,
    dense_rank,
    min_rank,
    ntile,
    percent_rank,
    row_number,
)
from .recode import recode, recode_categorical, recode_factor
from .relocate import relocate
from .rename import rename, rename_with
from .rows import rows_delete, rows_insert, rows_patch, rows_update, rows_upsert
from .select import select
from .sets import intersect, setdiff, setequal, union, union_all
from .summarise import summarise, summarize
from .tidyselect import (
    all_of,
    any_of,
    contains,
    ends_with,
    everything,
    last_col,
    matches,
    num_range,
    starts_with,
    where,
)

# make sure builtin names are included when
# from datar.dplyr import *
_builtin_names = {
    "filter": filter_,
    "slice": slice_,
}

__all__ = [var_ for var_ in locals() if not var_.startswith("_")]
__all__.extend(_builtin_names)

# warn when builtin names are imported directly
from ..core.warn_builtin_names import warn_builtin_names

__getattr__ = warn_builtin_names(**_builtin_names)



"""Vectorised if and multiple if-else

https://github.com/tidyverse/dplyr/blob/master/R/if_else.R
https://github.com/tidyverse/dplyr/blob/master/R/case_when.R
"""
from typing import Any, Iterable, Union

import numpy
from pandas import Series
from pipda import register_func

from ..core.contexts import Context
from ..core.types import is_iterable, is_scalar, is_null
from ..core.utils import Array
from ..base import NA


@register_func(None, context=Context.EVAL)
def if_else(
    condition: Union[bool, Iterable[bool]],
    true: Any,
    false: Any,
    missing: Any = None,
) -> numpy.ndarray:
    """Where condition is TRUE, the matching value from true, where it's FALSE,
    the matching value from false, otherwise missing.

    Args:
        condition: the conditions
        true: and
        false: Values to use for TRUE and FALSE values of condition.
            They must be either the same length as condition, or length 1.
        missing: If not None, will be used to replace missing values

    Returns:
        A series with values replaced.
    """
    condition = Array(condition)
    na_indexes = is_null(condition)
    condition[na_indexes] = False
    condition = condition.astype(bool)
    return case_when(
        na_indexes,
        missing,
        numpy.invert(condition),
        false,
        condition,
        true,
        True,
        missing,
    )


@register_func(None, context=Context.EVAL)
def case_when(*when_cases: Any) -> Series:
    """Vectorise multiple `if_else()` statements.

    Args:
        *when_cases: A even-size sequence, with 2n-th element values to match,
            and 2(n+1)-th element the values to replace.
            When matching value is True, then next value will be default to
            replace

    Returns:
        A series with values replaced
    """
    if not when_cases or len(when_cases) % 2 != 0:
        raise ValueError("Number of arguments of case_when should be even.")

    out_len = 1
    if is_iterable(when_cases[0]):
        out_len = len(when_cases[0])
    elif is_iterable(when_cases[1]):
        out_len = len(when_cases[1])

    out = Array([NA] * out_len, dtype=object)
    when_cases = reversed(list(zip(when_cases[0::2], when_cases[1::2])))
    for case, rep in when_cases:
        if case is True:
            out[:] = rep
        elif (
            case is not False and case is not NA
        ):  # skip atmoic False condition
            case = Array(case)
            case[is_null(case)] = False
            case = case.astype(bool)
            index = numpy.where(case)
            if is_scalar(rep) or len(rep) == 1 or len(rep) == len(index):
                out[index] = rep
            elif len(rep) == len(out):
                out[index] = Array(rep)[index]
            else:
                raise ValueError(
                    f"Values to replace must be length {out_len} "
                    f"(length of `condition`) or one, not {len(rep)}."
                )

    return Array(out.tolist())  # shrink the dtype



"""Functions related to randomization"""
import random as std_random

import numpy


def set_seed(seed: int) -> None:
    """Set the seed for randomization"""
    std_random.seed(seed)
    numpy.random.seed(seed)



"""Provide functions to manipulate multiple rows

https://github.com/tidyverse/dplyr/blob/master/R/rows.R
"""
from typing import List

import numpy
from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..base import setdiff
from ..core.types import StringOrIter, is_scalar, is_null, is_not_null
from ..core.utils import logger
from ..tibble import rownames_to_column

from .bind import bind_rows
from .join import left_join
from .funs import coalesce


@register_verb(DataFrame)
def rows_insert(
    x: DataFrame,
    y: DataFrame,
    by: StringOrIter = None,
    copy: bool = True,
) -> DataFrame:
    """Adds new rows to a data frame

    Argument `in_place` not supported, as we always do data frames here.

    Args:
        x: The seed data frame
        y: The data frame with rows to be inserted into `x`.
            - Key values in `y` must not occur in `x`
            - `y` must have the same or a subset columns of `x`
        by: A string or a list of strings giving the key columns.
            The key values must uniquely identify each row
            (i.e. each combination of key values occurs at most once),
            and the key columns must exist in both x and y.
            By default, we use the first column in y, since the first column
            is a reasonable place to put an identifier variable.
        copy: If `False`, do not copy data unnecessarily.
            Original API does not support this. This argument will be
            passed by to `pandas.concat()` as `copy` argument.

    Returns:
        A data frame with `y` inserted into `x`
    """
    key = _rows_check_key(by, x, y)
    _rows_check_key_df(x, key, df_name="x")
    _rows_check_key_df(y, key, df_name="y")

    idx = _rows_match(y[key], x[key])
    bad = is_not_null(idx)
    if any(bad):
        raise ValueError("Attempting to insert duplicate rows.")

    return bind_rows(x, y, _copy=copy, __calling_env=CallingEnvs.REGULAR)


@register_verb(DataFrame)
def rows_update(
    x: DataFrame,
    y: DataFrame,
    by: StringOrIter = None,
    copy: bool = True,
) -> DataFrame:
    """Modifies existing rows in a data frame

    See Also:
        [`rows_insert`](datar.dplyr.rows.rows_insert)

    Args:
        x: The seed data frame
        y: The data frame with rows to be inserted into `x`.
            - Key values in `y` must not occur in `x`
            - `y` must have the same or a subset columns of `x`
        by: A string or a list of strings giving the key columns.
            The key values must uniquely identify each row
            (i.e. each combination of key values occurs at most once),
            and the key columns must exist in both x and y.
            By default, we use the first column in y, since the first column
            is a reasonable place to put an identifier variable.
        copy: Whether `x` should be copied and updated or updated directly

    Returns:
        `x` with values of keys updated
    """
    key = _rows_check_key(by, x, y)
    _rows_check_key_df(x, key, df_name="x")
    _rows_check_key_df(y, key, df_name="y")

    idx = _rows_match(y[key], x[key])
    bad = is_null(idx)
    if any(bad):
        raise ValueError("Attempting to update missing rows.")

    idx = idx.astype(int)
    if copy:
        x = x.copy()
    x.loc[idx, y.columns] = y.values
    return x


@register_verb(DataFrame)
def rows_patch(
    x: DataFrame,
    y: DataFrame,
    by: StringOrIter = None,
    copy: bool = True,
) -> DataFrame:
    """Works like `rows_update()` but only overwrites `NA` values.

    See Also:
        [`rows_insert`](datar.dplyr.rows.rows_insert)

    Args:
        x: The seed data frame
        y: The data frame with rows to be inserted into `x`.
            - Key values in `y` must not occur in `x`
            - `y` must have the same or a subset columns of `x`
        by: A string or a list of strings giving the key columns.
            The key values must uniquely identify each row
            (i.e. each combination of key values occurs at most once),
            and the key columns must exist in both x and y.
            By default, we use the first column in y, since the first column
            is a reasonable place to put an identifier variable.
        copy: Whether `x` should be copied and updated or updated directly

    Returns:
        `x` with values of keys updated
    """
    key = _rows_check_key(by, x, y)
    _rows_check_key_df(x, key, df_name="x")
    _rows_check_key_df(y, key, df_name="y")

    idx = _rows_match(y[key], x[key])
    bad = is_null(idx)
    if any(bad):
        raise ValueError("Attempting to patch missing rows.")

    new_data = []
    for col in y.columns:
        new_data.append(coalesce(x.loc[idx, col].values, y[col]))

    if copy:
        x = x.copy()
    x.loc[idx, y.columns] = numpy.array(new_data).T
    return x


@register_verb(DataFrame)
def rows_upsert(
    x: DataFrame,
    y: DataFrame,
    by: StringOrIter = None,
    copy: bool = True,
) -> DataFrame:
    """Inserts or updates depending on whether or not the
    key value in `y` already exists in `x`.

    See Also:
        [`rows_insert`](datar.dplyr.rows.rows_insert)

    Args:
        x: The seed data frame
        y: The data frame with rows to be inserted into `x`.
            - Key values in `y` must not occur in `x`
            - `y` must have the same or a subset columns of `x`
        by: A string or a list of strings giving the key columns.
            The key values must uniquely identify each row
            (i.e. each combination of key values occurs at most once),
            and the key columns must exist in both x and y.
            By default, we use the first column in y, since the first column
            is a reasonable place to put an identifier variable.
        copy: If `False`, do not copy data unnecessarily.
            Original API does not support this. This argument will be
            passed by to `pandas.concat()` as `copy` argument.

    Returns:
        `x` with values of keys updated
    """
    key = _rows_check_key(by, x, y)
    _rows_check_key_df(x, key, df_name="x")
    _rows_check_key_df(y, key, df_name="y")

    idx = _rows_match(y[key], x[key])
    new = is_null(idx)
    # idx of x
    idx_existing = idx[~new]

    x.loc[idx_existing, y.columns] = y.loc[~new].values
    return bind_rows(
        x, y.loc[new], _copy=copy, __calling_env=CallingEnvs.REGULAR
    )


@register_verb(DataFrame)
def rows_delete(
    x: DataFrame,
    y: DataFrame,
    by: StringOrIter = None,
    copy: bool = True,
) -> DataFrame:
    """Deletes rows; key values in `y` must exist in `x`.

    See Also:
        [`rows_insert`](datar.dplyr.rows.rows_insert)

    Args:
        x: The seed data frame
        y: The data frame with rows to be inserted into `x`.
            - Key values in `y` must not occur in `x`
            - `y` must have the same or a subset columns of `x`
        by: A string or a list of strings giving the key columns.
            The key values must uniquely identify each row
            (i.e. each combination of key values occurs at most once),
            and the key columns must exist in both x and y.
            By default, we use the first column in y, since the first column
            is a reasonable place to put an identifier variable.
        copy: Whether `x` should be copied and deleted or deleted directly

    Returns:
        `x` with values of keys deleted
    """
    key = _rows_check_key(by, x, y)
    _rows_check_key_df(x, key, df_name="x")
    _rows_check_key_df(y, key, df_name="y")

    extra_cols = setdiff(y.columns, key, __calling_env=CallingEnvs.REGULAR)
    if len(extra_cols) > 0:
        logger.info("Ignoring extra columns: %s", extra_cols)

    idx = _rows_match(y[key], x[key])
    bad = is_null(idx)

    if any(bad):
        raise ValueError("Attempting to delete missing rows.")

    if copy:
        x = x.copy()

    return x.loc[~x.index.isin(idx), :]


# helpers -----------------------------------------------------------------


def _rows_check_key(by: StringOrIter, x: DataFrame, y: DataFrame) -> List[str]:
    """Check the key and return the valid key"""
    if by is None:
        by = y.columns[0]
        logger.info("Matching, by=%r", by)

    if is_scalar(by):
        by = [by]  # type: ignore

    for by_elem in by:
        if not isinstance(by_elem, str):
            raise ValueError("`by` must be a string or a list of strings.")

    bad = setdiff(y.columns, x.columns, __calling_env=CallingEnvs.REGULAR)
    if len(bad) > 0:
        raise ValueError("All columns in `y` must exist in `x`.")

    return by


def _rows_check_key_df(df: DataFrame, by: List[str], df_name: str) -> None:
    """Check key with the data frame"""
    y_miss = setdiff(by, df.columns, __calling_env=CallingEnvs.REGULAR)
    if len(y_miss) > 0:
        raise ValueError(f"All `by` columns must exist in `{df_name}`.")

    if any(df.duplicated(by)):
        raise ValueError(f"`{df_name}` key values are not unique.")


def _rows_match(x: DataFrame, y: DataFrame) -> numpy.ndarray:
    """Mimic vctrs::vec_match"""
    id_col = "__id__"
    y_with_id = rownames_to_column(
        y, var=id_col, __calling_env=CallingEnvs.REGULAR
    )

    return left_join(x, y_with_id, __calling_env=CallingEnvs.REGULAR)[
        id_col
    ].values



"""Provides desc"""
from typing import Any, Iterable
from pandas import Series, Categorical
from pipda import register_func

from ..core.contexts import Context
from ..base import NA


@register_func(None, context=Context.EVAL)
def desc(x: Iterable[Any]) -> Series:
    """Transform a vector into a format that will be sorted in descending order

    This is useful within arrange().

    The original API:
    https://dplyr.tidyverse.org/reference/desc.html

    Args:
        x: vector to transform

    Returns:
        The descending order of x
    """
    x = Series(x)
    try:
        return -x
    except TypeError:
        cat = Categorical(x)
        code = Series(cat.codes).astype(float)
        code[code == -1.0] = NA
        return -code



"""Subset columns using their names and types

See source https://github.com/tidyverse/dplyr/blob/master/R/select.R
"""
from typing import Any, Iterable, List, Mapping, Tuple, Union

from pandas import DataFrame, Index, Series
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.types import StringOrIter
from ..core.utils import position_at, vars_select, logger
from ..core.collections import Inverted
from ..core.grouped import DataFrameGroupBy
from ..base import setdiff, union
from .group_by import group_by_drop_default
from .group_data import group_data, group_vars


@register_verb(DataFrame, context=Context.SELECT)
def select(
    _data: DataFrame,
    *args: Union[StringOrIter, Iterable[Iterable], Inverted],
    base0_: bool = None,
    **kwargs: Mapping[str, str],
) -> DataFrame:
    """Select (and optionally rename) variables in a data frame

    See original API
    https://dplyr.tidyverse.org/reference/select.html

    To exclude columns use `~` instead of `-`. For example, to exclude last
    column: `select(df, ~c(-1))`.

    To use column name in slice: `f[f.col1:f.col2]`. If you don't want `col2`
    to be included: `f[f.col1:f.col2:0]`

    Args:
        *columns: The columns to select
        **renamings: The columns to rename and select in new => old column way.
        base0_: Whether the columns are 0-based if given by indexes
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The dataframe with select columns
    """
    all_columns = _data.columns
    gvars = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
    selected, new_names = _eval_select(
        all_columns,
        *args,
        **kwargs,
        _group_vars=gvars,
        base0_=base0_,
    )
    out = _data.iloc[:, selected].copy()

    if new_names:
        out.rename(columns=new_names, inplace=True)

    if isinstance(_data, DataFrameGroupBy):
        gvars = [new_names.get(gvar, gvar) for gvar in gvars]
        gdata = group_data(_data, __calling_env=CallingEnvs.REGULAR)
        gdata.columns = gvars + ["_rows"]
        return _data.__class__(
            out,
            _group_vars=gvars,
            _group_drop=group_by_drop_default(_data),
            _group_data=gdata,
        )
    return out


def _eval_select(
    _all_columns: Index,
    *args: Any,
    _group_vars: Iterable[str],
    base0_: bool = None,
    **kwargs: Any,
) -> Tuple[List[int], Mapping[str, str]]:
    """Evaluate selections to get locations

    Returns:
        A tuple of (selected columns, dict of old-to-new renaming columns)
    """
    selected = vars_select(
        _all_columns,
        *args,
        *kwargs.values(),
        base0=base0_,
    )
    missing = setdiff(
        _group_vars, _all_columns[selected], __calling_env=CallingEnvs.REGULAR
    )
    if missing:
        logger.info("Adding missing grouping variables: %s", missing)

    selected = union(
        _all_columns.get_indexer_for(_group_vars),
        selected,
        __calling_env=CallingEnvs.REGULAR,
    )

    # dplyr takes new -> old
    # we transform it to old -> new for better access
    new_names = {}
    for key, val in kwargs.items():
        # key: new name
        # val: old name
        if isinstance(val, Series):
            val = val.name
        if isinstance(val, str):
            idx = _all_columns.get_indexer_for([val])
            if len(idx) > 1:
                raise ValueError(
                    "Names must be unique. "
                    f'Name "{val}" found at locations {list(idx)}.'
                )
        else:  # int
            # try:
            #   If out of bounds, it should be raised at getting missing
            val = _all_columns[
                position_at(val, len(_all_columns), base0=base0_)
            ]
            # except IndexError:
            #     raise ColumnNotExistingError(
            #         f'Location {val} does not exist.'
            #     ) from None
        new_names[val] = key
    return selected, new_names



"""Windowed rank functions.

See https://github.com/tidyverse/dplyr/blob/master/R/rank.R
"""
from typing import Any, Iterable

import numpy
import pandas
from pandas.core.frame import DataFrame
from pipda import register_func

from ..core.contexts import Context
from ..core.types import is_null
from ..core.utils import Array
from ..base import NA


@register_func(context=Context.EVAL)
def row_number(_data: Iterable[Any], x: Iterable[Any] = None) -> numpy.ndarray:
    """Gives the row number

    See https://dplyr.tidyverse.org/reference/ranking.html

    Args:
        x: a vector of values to rank

    Returns:
        The row number of `x` or the data frame (1-based)
    """
    if x is None:
        return Array(range(len(_data))) + 1
    return _rank(x, na_last="keep", method="first")


@register_func(context=Context.EVAL)
def ntile(
    _data: Iterable[Any],
    series: Iterable[Any] = None,
    n: int = None,
) -> Iterable[Any]:
    """A rough rank, which breaks the input vector into ‘n’ buckets.

    Note:
        The output tiles are 0-based.
        The result is slightly different from dplyr's ntile.
        >>> ntile(c(1,2,NA,1,0,NA), 2) # dplyr
        >>> # 1 2 NA 2 1 NA
        >>> ntile([1,2,NA,1,0,NA], n=2) # datar
        >>> # [0, 1, NA, 0, 0, NA]
        >>> # Categories (2, int64): [0 < 1]
    """
    if isinstance(series, int) and n is None:
        n = series
        series = None

    if series is None:
        series = row_number(_data) if isinstance(_data, DataFrame) else _data

    # if is_scalar(series):
    #     series = [series]
    # support generator
    series = list(series)

    if len(series) == 0:
        return pandas.Categorical([])
    if all(is_null(series)):
        return pandas.Categorical([NA] * len(series))
    n = min(n, len(series))
    return pandas.cut(series, n, labels=range(n))


@register_func(None, context=Context.EVAL)
def min_rank(series: Iterable[Any], na_last: str = "keep") -> Iterable[float]:
    """Rank the data using min method"""
    return _rank(series, na_last=na_last, method="min")


@register_func(None, context=Context.EVAL)
def dense_rank(series: Iterable[Any], na_last: str = "keep") -> Iterable[float]:
    """Rank the data using dense method"""
    return _rank(series, na_last=na_last, method="dense")


@register_func(None, context=Context.EVAL)
def percent_rank(
    series: Iterable[Any], na_last: str = "keep"
) -> Iterable[float]:
    """Rank the data using percent_rank method"""
    # Without this:
    # pandas 1.2.0: ValueError: Transform function failed
    if len(series) == 0:
        dtype = getattr(series, "dtype", None)
        return Array(series, dtype=dtype)  # make a copy

    ranking = _rank(series, na_last, "min", True)
    minrank = ranking.min()
    maxrank = ranking.max()
    ret = ranking.transform(lambda r: (r - minrank) / (maxrank - minrank))
    ret[ranking.isna()] = NA
    return ret


@register_func(None, context=Context.EVAL)
def cume_dist(series: Iterable[Any], na_last: str = "keep") -> Iterable[float]:
    """Rank the data using percent_rank method"""
    # Without this:
    # pandas 1.2.0: ValueError: Transform function failed
    if len(series) == 0:
        dtype = getattr(series, "dtype", None)
        return Array(series, dtype=dtype)  # make a copy

    ranking = _rank(series, na_last, "min")
    max_ranking = ranking.max()
    ret = ranking.transform(lambda r: ranking.le(r).sum() / max_ranking)
    ret[ranking.isna()] = NA
    return ret


def _rank(
    data: Iterable[Any], na_last: str, method: str, percent: bool = False
) -> Iterable[float]:
    """Rank the data"""
    if not isinstance(data, pandas.Series):
        data = pandas.Series(data)

    # ascending = not isinstance(data, DescSeries)
    ascending = True

    ret = data.rank(
        method=method,
        ascending=ascending,
        pct=percent,
        na_option=(
            "keep" if na_last == "keep" else "top" if not na_last else "bottom"
        ),
    )
    return ret



"""Summarise each group to fewer rows"""

from typing import Any, Iterable, Mapping, Union

from pandas import DataFrame, Series
from pipda import register_verb, evaluate_expr
from pipda.function import Function
from pipda.utils import CallingEnvs

from ..core.defaults import DEFAULT_COLUMN_PREFIX
from ..core.contexts import Context
from ..core.utils import (
    length_of,
    recycle_df,
    arg_match,
    check_column_uniqueness,
    df_setitem,
    name_mutatable_args,
    logger,
    get_option,
    reconstruct_tibble,
    dedup_name,
)
from ..core.exceptions import ColumnNotExistingError
from ..core.grouped import DataFrameGroupBy, DataFrameRowwise

from .group_data import group_keys, group_vars, group_data
from .group_by import group_by_drop_default


@register_verb(DataFrame, context=Context.PENDING)
def summarise(
    _data: DataFrame,
    *args: Union[DataFrame, Mapping[str, Iterable[Any]]],
    _groups: str = None,
    **kwargs: Any,
) -> DataFrame:
    """Summarise each group to fewer rows

    See https://dplyr.tidyverse.org/reference/summarise.html

    Both input and the summarised data can be recycled, but separately.

    Aliases - `summarize`

    Examples:
        >>> df = tibble(x=[1,2,3,4])
        >>> df >> summarise(y=sum(f.x), z=f.y*2)
        >>> #   y  z
        >>> # 0 10 20
        >>> df >> summarise(y=sum(f.x), z=f.x+f.y) # fail

        But they should not be mixed in later argument. For example:
        >>> df = tibble(x=[1,2,3,4], g=list('aabb')) >> group_by(f.g)
        >>> df >> summarise(n=n() + f.x)
        >>> # as expected:
        >>>      g  n
        >>> # 0  a  3
        >>> # 1  a  4
        >>> # 2  b  5
        >>> # 3  b  6
        >>> # [Groups: ['g'] (n=2)]
        >>> # However:
        >>> df >> summarise(y=1, n=n() + f.y)
        >>> # n() will be recycling output instead of input
        >>> #    g  y  n
        >>> # 0  a  1  2
        >>> # 1  b  1  2

    Args:
        _groups: Grouping structure of the result.
            - "drop_last": dropping the last level of grouping.
            - "drop": All levels of grouping are dropped.
            - "keep": Same grouping structure as _data.
            - "rowwise": Each row is its own group.
        *args: and
        **kwargs: Name-value pairs, where value is the summarized
            data for each group

    Returns:
        The summary dataframe.
    """
    check_column_uniqueness(
        _data, "Can't transform a data frame with duplicate names"
    )
    _groups = arg_match(
        _groups, "_groups", ["drop", "drop_last", "keep", "rowwise", None]
    )
    out = _summarise_build(_data, *args, **kwargs)
    if _groups == "rowwise":
        return DataFrameRowwise(out, _group_drop=group_by_drop_default(_data))
    return out


@summarise.register(DataFrameGroupBy, context=Context.PENDING)
def _(
    _data: DataFrameGroupBy,
    *args: Union[DataFrame, Mapping[str, Iterable[Any]]],
    _groups: str = None,
    **kwargs: Any,
) -> DataFrame:
    # empty
    _groups = arg_match(
        _groups, "_groups", ["drop", "drop_last", "keep", "rowwise", None]
    )

    allone = True
    if group_data(_data, __calling_env=CallingEnvs.REGULAR).shape[0] == 0:
        out = _summarise_build(_data, *args, **kwargs).iloc[[], :]
    else:

        def apply_func(df):
            nonlocal allone
            if isinstance(df, Series):
                # rowwise
                df.name = 0
                df = df.to_frame().T
            out = summarise(
                df, *args, **kwargs, __calling_env=CallingEnvs.REGULAR
            )
            if out.shape[0] != 1:
                allone = False
            return out

        mappings = {}
        for i, arg in enumerate(args):
            if isinstance(arg, dict):
                mappings.update(arg)
            else:
                mappings[f"{DEFAULT_COLUMN_PREFIX}{i}"] = arg
        mappings.update(kwargs)

        out = _data._datar_apply(apply_func, _mappings=mappings, _method="agg")

    g_keys = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
    if _groups is None:
        if allone and not isinstance(_data, DataFrameRowwise):
            _groups = "drop_last"
        else:
            _groups = "keep"

    if _groups == "drop_last":
        if len(g_keys) > 1:
            if get_option("dplyr_summarise_inform"):
                logger.info(
                    "`summarise()` has grouped output by "
                    "%s (override with `_groups` argument)",
                    g_keys[:-1],
                )
            out = reconstruct_tibble(_data, out, [g_keys[-1]])
    elif _groups == "keep" and g_keys:
        if get_option("dplyr_summarise_inform"):
            logger.info(
                "`summarise()` has grouped output by "
                "%s (override with `_groups` argument)",
                g_keys,
            )
        out = DataFrameGroupBy(
            out,
            _group_vars=g_keys,
            _group_drop=group_by_drop_default(_data),
        )
    elif _groups == "rowwise":
        out = reconstruct_tibble(
            _data,
            out,
            keep_rowwise=True,
        )
    elif isinstance(_data, DataFrameRowwise) and get_option(
        "dplyr_summarise_inform"
    ):
        logger.info(
            "`summarise()` has ungrouped output. "
            "You can override using the `_groups` argument."
        )
    # else: # drop
    return out


summarize = summarise


def _summarise_build(_data: DataFrame, *args: Any, **kwargs: Any) -> DataFrame:
    """Build summarise result"""
    context = Context.EVAL.value
    named = name_mutatable_args(*args, **kwargs)

    out = group_keys(_data, __calling_env=CallingEnvs.REGULAR)
    for key, val in named.items():
        # support: df %>% muate(a=1, a=a+1)
        key = dedup_name(key, list(named))

        envdata = out
        if out.shape[1] == 0 or (
            isinstance(val, Function)
            and getattr(val._pipda_func, "summarise_prefers_input", False)
        ):
            envdata = _data

        try:
            val = evaluate_expr(val, envdata, context)
        except (KeyError, ColumnNotExistingError):
            # also recycle input
            val = evaluate_expr(val, _data, context)

        if val is None:
            continue

        if key.startswith(DEFAULT_COLUMN_PREFIX) and isinstance(val, DataFrame):
            # ignore key
            for name, ser in val.to_dict("series").items():
                if length_of(out) == 1:
                    out, ser = recycle_df(out, ser, None, key)
                out = df_setitem(out, name, ser)
        elif isinstance(val, DataFrame):
            for name, ser in val.to_dict("series").items():
                if length_of(out) == 1:
                    out, ser = recycle_df(out, ser, None, key)
                out = df_setitem(out, f"{key}${name}", ser)
        elif length_of(out) == 1:
            out, val = recycle_df(out, val, None, key)
            out = df_setitem(out, key, val)
        else:
            out = df_setitem(out, key, val)

    return out



"""Core utilities"""

import sys
import logging
import inspect
from functools import singledispatch
from copy import deepcopy
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Sequence,
    Union,
    Tuple,
)

import numpy
from numpy import array as Array

import pandas
from pandas import Categorical, DataFrame, Series
from pipda import register_func
from pipda.symbolic import Reference
from pipda.utils import CallingEnvs

from .exceptions import (
    ColumnNotExistingError,
    DataUnrecyclable,
    NameNonUniqueError,
)
from .contexts import Context
from .types import (
    StringOrIter,
    Dtype,
    is_iterable,
    is_scalar,
    is_categorical,
    is_null,
)
from .defaults import DEFAULT_COLUMN_PREFIX, NA_REPR

# logger
logger = logging.getLogger("datar")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler(sys.stderr)
stream_handler.setFormatter(
    logging.Formatter(
        "[%(asctime)s][%(name)s][%(levelname)7s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
logger.addHandler(stream_handler)


def vars_select(
    all_columns: Iterable[str],
    *columns: Any,
    raise_nonexists: bool = True,
    base0: bool = None,
) -> List[int]:
    # TODO: support selecting data-frame columns
    """Select columns

    Args:
        all_columns: The column pool to select
        *columns: arguments to select from the pool
        raise_nonexist: Whether raise exception when column not exists
            in the pool
        base0: Whether indexes are 0-based if columns are selected by indexes.
            If not given, will use `datar.base.get_option('index.base.0')`

    Returns:
        The selected indexes for columns

    Raises:
        ColumnNotExistingError: When the column does not exist in the pool
            and raise_nonexists is True.
    """
    from .collections import Collection
    from ..base import unique

    columns = [
        column.name if isinstance(column, Series) else column
        for column in columns
    ]
    selected = Collection(*columns, pool=list(all_columns), base0=base0)
    if raise_nonexists and selected.unmatched and selected.unmatched != {None}:
        raise ColumnNotExistingError(
            f"Columns `{selected.unmatched}` do not exist."
        )
    return unique(selected).astype(int)


def recycle_value(
    value: Any, size: int, name: str = None
) -> Union[DataFrame, numpy.ndarray]:
    """Recycle a value based on a dataframe

    Args:
        value: The value to be recycled
        size: The size to recycle to
        name: The name to show in the error if failed to recycle

    Returns:
        The recycled value
    """
    # TODO: follow base R's recycling rule? i.e. size 2 -> 4
    from ..base import NA

    if is_scalar(value):
        value = [value]

    length = len(value)

    if length not in (0, 1, size):
        name = "value" if not name else f"`{name}`"
        expect = "1" if size == 1 else f"(1, {size})"
        raise DataUnrecyclable(
            f"Cannot recycle {name} to size {size}, "
            f"expect {expect}, got {length}."
        )

    if isinstance(value, DataFrame):
        if length == size == 0:
            return DataFrame(columns=value.columns)
        if length == 0:
            value = DataFrame([[NA] * value.shape[1]], columns=value.columns)
        if length == 1 and size > length:
            return value.iloc[[0] * size, :].reset_index(drop=True)
        return value

    cats = categorized(value).categories if is_categorical(value) else None

    if length == size == 0:
        return [] if cats is None else Categorical([], categories=cats)

    if length == 0:
        value = [NA]

    if isinstance(value, Series):
        # try to keep Series class
        # some operators can only do with it or with it correctly
        # For example:
        # Series([True, True]) & Series([False, NA]) -> [False, Fa.se]
        # But with numpy.array, it raises error, since NA is a float
        if length == 1 and size > length:
            value = value.iloc[[0] * size].reset_index(drop=True)
        return value

    if isinstance(value, tuple):
        value = list(value)

    # dtype = getattr(value, 'dtype', None)
    if length == 1 and size > length:
        value = list(value) * size

    if cats is not None:
        return Categorical(value, categories=cats)

    is_elem_iter = any(is_iterable(val) for val in value)
    if is_elem_iter:
        # without dtype: VisibleDeprecationWarning
        # return Array(value, dtype=object)
        # The above does not keep [DataFrame()] structure
        return value

    # Avoid numpy.nan to be converted into 'nan' when other elements are string
    out = Array(value)
    if numpy.issubdtype(out.dtype, numpy.str_) and is_null(value).any():
        return Array(value, dtype=object)
    return out


def recycle_df(
    df: DataFrame,
    value: Any,
    df_name: str = None,
    value_name: str = None,
) -> Tuple[DataFrame, Any]:
    """Recycle the dataframe based on value"""
    if length_of(df) == 1:
        df = recycle_value(df, length_of(value), df_name)
    value = recycle_value(value, length_of(df), value_name)
    return df, value


def categorized(data: Any) -> Any:
    """Get the Categorical object"""
    if not is_categorical(data):
        return data
    if isinstance(data, Series):
        return data.values
    return data


@singledispatch
def to_df(data: Any, name: str = None) -> DataFrame:
    """Convert an object to a data frame"""
    if is_scalar(data):
        data = [data]

    if name is None:
        return DataFrame(data)

    return DataFrame({name: data})


@to_df.register(numpy.ndarray)
def _(data: numpy.ndarray, name: StringOrIter = None) -> DataFrame:
    if name is not None and is_scalar(name):
        name = [name]

    if len(data.shape) == 1:
        return (
            DataFrame(data, columns=name)
            if name is not None
            else DataFrame(data)
        )

    ncols = data.shape[1]
    if name is not None and len(name) == ncols:
        return DataFrame(data, columns=name)
    # ignore the name
    return DataFrame(data)


@to_df.register(DataFrame)
def _(data: DataFrame, name: str = None) -> DataFrame:
    if name is None:
        return data
    return DataFrame({f"{name}${col}": data[col] for col in data.columns})


@to_df.register(Series)
def _(data: Series, name: str = None) -> DataFrame:
    name = name or data.name
    return data.to_frame(name=name)


# @to_df.register(SeriesGroupBy)
# def _(data: SeriesGroupBy, name: str = None) -> DataFrame:
#     name = name or data.obj.name
#     return data.obj.to_frame(name=name).groupby(data.grouper, dropna=False)


def check_column_uniqueness(df: DataFrame, msg: str = None) -> None:
    """Check if column names are unique of a dataframe"""
    uniq = set()
    for col in df.columns:
        if col not in uniq:
            uniq.add(col)
        else:
            msg = msg or "Name is not unique"
            raise NameNonUniqueError(f"{msg}: {col}")


def dict_insert_at(
    container: Mapping[str, Any],
    poskeys: Sequence[str],
    value: Mapping[str, Any],
    remove: bool = False,
) -> Mapping[str, Any]:
    """Insert value to a certain position of a dict"""
    ret_items = []  # type: List[Tuple[str, Any]]
    ret_items_append = ret_items.append
    matched = False
    for key, val in container.items():
        if key == poskeys[0]:
            matched = True
            if not remove:
                ret_items_append((key, val))
            ret_items.extend(value.items())
        elif matched and key in poskeys:
            if not remove:
                ret_items_append((key, val))
        elif matched and key not in poskeys:
            matched = False
            ret_items_append((key, val))
        else:
            ret_items_append((key, val))

    return dict(ret_items)


def name_mutatable_args(
    *args: Union[Series, DataFrame, Mapping[str, Any]],
    **kwargs: Any,
) -> Mapping[str, Any]:
    """Convert all mutatable arguments to named mappings, which can be easier
    to mutate later on.

    If there are Expression objects, keep it. So if an objects have multiple
    names and it's built by an Expression, then the name might get lost here.

    Examples:

        >>> s = Series([1], name='a')
        >>> name_mutatable_args(s, b=2)
        >>> # {'a': s, 'b': 2}
        >>> df = DataFrame({'x': [3], 'y': [4]})
        >>> name_mutatable_args(df)
        >>> # {'x': [3], 'y': [4]}
        >>> name_mutatable_args(d=df)
        >>> # {'d$x': [3], 'd$y': [4]}
    """
    # order kept
    ret = {}  # type: dict

    for i, arg in enumerate(args):
        if isinstance(arg, Series):
            ret[arg.name] = arg
        elif isinstance(arg, dict):
            ret.update(arg)
        elif isinstance(arg, DataFrame):
            ret.update(arg.to_dict("series"))
        elif isinstance(arg, Reference):
            ret[arg._pipda_ref] = arg
        else:
            ret[f"{DEFAULT_COLUMN_PREFIX}{i}"] = arg

    for key, val in kwargs.items():
        if isinstance(val, DataFrame):
            val = val.to_dict("series")

        if isinstance(val, dict):
            existing_keys = [
                ret_key
                for ret_key in ret
                if ret_key == key or ret_key.startswith(f"{key}$")
            ]
            if existing_keys:
                ret = dict_insert_at(  # type: ignore
                    ret, existing_keys, {key: val}, remove=True
                )
            else:
                for dkey, dval in val.items():
                    ret[f"{key}${dkey}"] = dval
        else:
            ret[key] = val
    return ret


def arg_match(
    arg: Any, argname: str, values: Iterable[Any], errmsg: str = None
) -> Any:
    """Make sure arg is in one of the values.

    Mimics `rlang::arg_match`.
    """
    if not errmsg:
        values = list(values)
        errmsg = f"`{argname}` must be one of {values}."
    if arg not in values:
        raise ValueError(errmsg)
    return arg


def copy_attrs(df1: DataFrame, df2: DataFrame, deep: bool = True) -> None:
    """Copy attrs from df2 to df1"""
    for key, val in df2.attrs.items():
        if key.startswith("_"):
            continue
        df1.attrs[key] = deepcopy(val) if deep else val


def nargs(fun: Callable) -> int:
    """Get the number of arguments of a function"""
    return len(inspect.signature(fun).parameters)


def position_at(
    pos: int, length: int, base0: bool = None, raise_exc: bool = True
) -> int:
    """Get the 0-based position right at the given pos

    When `pos` is negative, it acts like 0-based, meaning `-1` will anyway
    represent the last position regardless of `base0`

    Args:
        pos: The given position
        length: The length of the pool
        base0: Whether the given `pos` is 0-based
        raise_exc: Raise error if `pos` is out of range?

    Returns:
        The 0-based position
    """
    from .collections import Collection

    coll = Collection(pos, pool=length, base0=base0)
    if raise_exc and coll.error:

        raise coll.error
    return coll[0]


def position_after(pos: int, length: int, base0: bool = None) -> int:
    """Get the 0-based position right at the given pos

    Args:
        pos: The given position
        length: The length of the pool

    Returns:
        The position before the given position
    """
    base0 = get_option("index.base.0", base0)
    # after 0 with 1-based, should insert to first column
    if not base0 and pos == 0:
        return 0

    return position_at(pos, length, base0) + 1


def get_option(key: str, value: Any = None) -> Any:
    """Get the option with key.

    This is for interal use mostly.

    This is a shortcut for:
    >>> if value is not None:
    >>>     return value
    >>> from datar.base import get_option
    >>> return get_option(key)
    """
    if value is not None:
        return value
    from ..base import get_option as get_option_

    return get_option_(key)


def apply_dtypes(
    df: DataFrame, dtypes: Union[bool, Dtype, Mapping[str, Dtype]]
) -> None:
    """Apply dtypes to data frame"""
    if dtypes is None or dtypes is False:
        return

    if dtypes is True:
        inferred = df.convert_dtypes()
        for col in df:
            df[col] = inferred[col]
        return

    if not isinstance(dtypes, dict):
        dtypes = dict(zip(df.columns, [dtypes] * df.shape[1]))  # type: ignore

    for column, dtype in dtypes.items():
        if column in df:
            df[column] = df[column].astype(dtype)
        else:
            for col in df:
                if col.startswith(f"{column}$"):
                    df[col] = df[col].astype(dtype)


def keep_column_order(df: DataFrame, order: Iterable[str]):
    """Keep the order of columns as given `order`

    We cannot do `df[order]` directly, since `df` may have nested df columns.
    """
    out_columns = []
    for col in order:
        if col in df:
            out_columns.append(col)
        else:
            out_columns.extend(
                (dfcol for dfcol in df.columns if dfcol.startswith(f"{col}$"))
            )
    if set(out_columns) != set(df.columns):
        raise ValueError("Given `order` does not select all columns.")

    return df[out_columns]


def reconstruct_tibble(
    input: DataFrame,
    output: DataFrame,
    ungrouped_vars: List[str] = None,
    keep_rowwise: bool = False,
) -> DataFrame:
    """Reconstruct the output dataframe based on input dataframe

    Args:
        input: The input data frame
        output: The output data frame
        ungrouped_vars: Variables to exclude from grouping
        keep_rowwise: Whether rowwise structure should be kept

    Return:
        The reconstructed dataframe.
    """
    from ..base import setdiff, intersect
    from ..dplyr import group_vars, group_by_drop_default
    from .grouped import DataFrameGroupBy, DataFrameRowwise

    if ungrouped_vars is None:
        ungrouped_vars = []
    old_groups = group_vars(input, __calling_env=CallingEnvs.REGULAR)
    new_groups = intersect(
        setdiff(old_groups, ungrouped_vars, __calling_env=CallingEnvs.REGULAR),
        output.columns,
        __calling_env=CallingEnvs.REGULAR
    )

    if isinstance(input, DataFrameRowwise):
        out = (
            DataFrameRowwise(
                output,
                _group_vars=new_groups,
                _group_drop=group_by_drop_default(input),
            )
            if keep_rowwise
            else output
        )
    elif isinstance(input, DataFrameGroupBy) and len(new_groups) > 0:
        out = DataFrameGroupBy(
            output,
            _group_vars=new_groups,
            _group_drop=group_by_drop_default(input),
        )
    else:
        out = output

    copy_attrs(out, input)
    return out


def df_getitem(df: DataFrame, ref: Any) -> Union[DataFrame, numpy.ndarray]:
    """Select columns from a data frame

    If the column is a data frame, select that data frame.
    """
    try:
        return df[ref]
    except KeyError:
        cols = [col for col in df.columns if col.startswith(f"{ref}$")]
        if not cols:
            raise KeyError(ref) from None
        ret = df.loc[:, cols]
        ret.columns = [col[len(ref) + 1 :] for col in cols]
        return ret


def df_setitem(
    df: DataFrame, name: str, value: Any, allow_dups: bool = False
) -> DataFrame:
    """Assign an item to a dataframe

    Args:
        df: The data frame
        name: The name of the item
        value: The value to insert
        allow_dups: Allow duplicated names

    Returns:
        df itself or a merged df
    """
    value = recycle_value(value, df.shape[0])
    if isinstance(value, DataFrame):
        # nested df
        value.columns = [f"{name}${col}" for col in value.columns]

        if allow_dups:
            return pandas.concat([df, value], axis=1)

        for col in value.columns:
            df[col] = value[col]
        return df

    if isinstance(value, numpy.ndarray) and value.ndim > 1:
        # keep the list structure
        # otherwise, keep array to have dtype kept
        value = value.tolist()

    # drop the index/match the index of df
    if isinstance(value, Series):
        value = value.values

    # tuple turned into list in recycle_value
    # if isinstance(value, tuple):
    #     # ('A', array([1,2,3]))
    #     # VisibleDeprecationWarning
    #     value = list(value)

    if not allow_dups:
        df[name] = value

    else:
        df.insert(df.shape[1], name, value, allow_duplicates=True)

    return df


def fillna_safe(data: Iterable, rep: Any = NA_REPR) -> Iterable:
    """Safely replace NA in data, as we can't just fillna for
    a Categorical data directly.

    Args:
        data: The data to fill NAs
        rep: The replacement for NAs

    Returns:
        Data itself if it has no NAs.
        For Categorical data, `rep` will be added to categories
        Otherwise a Array object with NAs replaced with `rep` is returned with
        dtype object.

    Raises:
        ValueError: when `rep` exists in data
    """
    # elementwise comparison failed; returning scalar instead
    # if rep in data:
    if rep in list(data):
        raise ValueError("The value to replace NAs is already present in data.")

    if not is_null(data).any():
        return data

    if is_categorical(data):
        data = categorized(data)
        data = data.add_categories(rep)
        return data.fillna(rep)

    # rep may not be the same dtype as data
    return Series(data).fillna(rep).values


def na_if_safe(
    data: Iterable, value: str = NA_REPR, dtype: Dtype = None
) -> Iterable:
    """Replace value with NA

    Args:
        data: The data to replace value with NA
        value: The value to match
        dtype: Convert the data to dtype after replacement
            If data is Categorical, this is ignored

    Return:
        Array or categorical
    """
    if is_categorical(data):
        # ignore dtype
        data = categorized(data)
        # value to NA, value in categories removed
        return data.replace(value, None)

    from ..base import NA

    out = Series(data).replace(value, NA)
    return out if dtype is None else out.astype(dtype)


def length_of(x: Any) -> int:
    """Get the length of a value, scalar gets 1"""
    if is_scalar(x):
        return 1

    return len(x)


def dedup_name(name: str, all_names: Iterable[str]):
    """Check if a name is a duplicated name in all_names,
    return the deduplicated name.

    In other to support duplicated keyword arguments in R:
        >>> df %>% mutate(a=1, a=a*2)

    Now you can to it with datar:
        >>> df >> mutate(a_=1, a=f.a*2)

    Args:
        name: The name to deduplicate
        all_names: all names

    Returns:
        The deduplicated name
    """
    if not name.endswith("_") or name[:-1] not in all_names:
        return name

    # now determine whehter the real name is name[:-1]
    # because the name could be "a__" ("a_" is for sure in all_names)
    # we need to check if "a" is also in all_names
    # otherwise, the realname is "a_"
    name = name[:-1]
    while name.endswith("_") and name[:-1] in all_names:
        name = name[:-1]
    return name


def register_numpy_func_x(
    name: str,
    np_name: str,
    trans_in: Callable = None,
    trans_out: Callable = None,
    doc: str = "",
) -> Callable:
    """Register numpy function with single argument x

    Args:
        name: The name of the function to be returned
        np_name: The name of the function from numpy
        trans_in: Transformation for input before sending to numpy function
        trans_out: Transformation for output
        doc: The docstring for the function

    Returns:
        The registered function
    """
    func = getattr(numpy, np_name)

    @register_func(None, context=Context.EVAL)
    def _func(x: Any) -> Any:
        """Registered function from numpy"""
        if trans_in:
            x = trans_in(x)

        out = func(x)
        if trans_out:  # pragma: no cover
            out = trans_out(out)
        return out

    _func.__name__ = name
    _func.__doc__ = doc
    return _func



# BSD 3-Clause License

# Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc.
# and PyData Development Team
# All rights reserved.

# Copyright (c) 2011-2021, Open source contributors.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Monkey-patch data frame format to
1. add dtypes next to column names when printing
2. collapse data frames when they are elements of a parent data frame.
"""
# flake8: noqa
# code grabbed from pandas v1.2.5
# works with v1.3.3.
from pandas import DataFrame
from pandas.io.formats import format as fmt, html, string as stringf
from pandas.io.formats.html import (
    HTMLFormatter,
    NotebookFormatter,
    Mapping,
    MultiIndex,
    get_level_lengths,
)
from pandas.io.formats.format import (
    DataFrameFormatter,
    GenericArrayFormatter,
    partial,
    List,
    QUOTE_NONE,
    get_option,
    NA,
    NaT,
    np,
    PandasObject,
    extract_array,
    lib,
    notna,
    is_float,
    format_array,
)
from pandas.io.formats.string import StringFormatter
from pandas.io.formats.printing import pprint_thing
from pandas.core.dtypes.common import is_scalar
from pandas.core.dtypes.missing import isna

from .options import add_option


# TODO: patch more formatters

# pandas 1.2.0 doesn't have this function
def _trim_zeros_single_float(str_float: str) -> str:  # pragma: no cover
    """
    Trims trailing zeros after a decimal point,
    leaving just one if necessary.
    """
    str_float = str_float.rstrip("0")
    if str_float.endswith("."):
        str_float += "0"

    return str_float


class DatarDataFrameFormatter(DataFrameFormatter):  # pragma: no cover
    """Custom formatter for DataFrame"""

    @property
    def grouping_info(self) -> str:
        """Get the string representation of grouping info"""
        from .grouped import DataFrameGroupBy, DataFrameRowwise

        if isinstance(self.frame, DataFrameRowwise):
            return f"\n[Rowwise: {', '.join(self.frame.attrs['_group_vars'])}]"

        if isinstance(self.frame, DataFrameGroupBy):
            ngroups = self.frame._group_data.shape[0]
            return (
                f"\n[Groups: {', '.join(self.frame.attrs['_group_vars'])} "
                f"(n={ngroups})]"
            )

        return None

    def get_strcols(self) -> List[List[str]]:
        """
        Render a DataFrame to a list of columns (as lists of strings).
        """
        strcols = self._get_strcols_without_index()

        if self.index:
            #           dtype
            str_index = [""] + self._get_formatted_index(self.tr_frame)
            strcols.insert(0, str_index)

        return strcols

    def format_col(self, i: int) -> List[str]:
        """Format column, add dtype ahead"""
        frame = self.tr_frame
        formatter = self._get_formatter(i)
        dtype = frame.iloc[:, i].dtype.name

        return [f"<{dtype}>"] + format_array(
            frame.iloc[:, i]._values,
            formatter,
            float_format=self.float_format,
            na_rep=self.na_rep,
            space=self.col_space.get(frame.columns[i]),
            decimal=self.decimal,
            leading_space=self.index,
        )


class DatarGenericArrayFormatter(GenericArrayFormatter):  # pragma: no cover
    """Generic Array Formatter to show DataFrame element in a cell in a
    collpased representation
    """

    def _format_strings(self) -> List[str]:
        if self.float_format is None:
            float_format = get_option("display.float_format")
            if float_format is None:
                precision = get_option("display.precision")
                float_format = lambda x: f"{x: .{precision:d}f}"
        else:
            float_format = self.float_format

        if self.formatter is not None:
            formatter = self.formatter
        else:
            quote_strings = (
                self.quoting is not None and self.quoting != QUOTE_NONE
            )
            formatter = partial(
                pprint_thing,
                escape_chars=("\t", "\r", "\n"),
                quote_strings=quote_strings,
            )

        def _format(x):
            if self.na_rep is not None and is_scalar(x) and isna(x):
                try:
                    # try block for np.isnat specifically
                    # determine na_rep if x is None or NaT-like
                    if x is None:
                        return "None"
                    if x is NA:
                        return str(NA)
                    if x is NaT or np.isnat(x):
                        return "NaT"
                except (TypeError, ValueError):
                    # np.isnat only handles datetime or timedelta objects
                    pass
                return self.na_rep
            # Show data frame as collapsed representation
            if isinstance(x, DataFrame):
                return f"<DF {x.shape[0]}x{x.shape[1]}>"
            if isinstance(x, PandasObject):
                return str(x)
            # else:
            # object dtype
            return str(formatter(x))

        vals = extract_array(self.values, extract_numpy=True)

        is_float_type = (
            lib.map_infer(vals, is_float)
            # vals may have 2 or more dimensions
            & np.all(notna(vals), axis=tuple(range(1, len(vals.shape))))
        )
        leading_space = self.leading_space
        if leading_space is None:
            leading_space = is_float_type.any()

        fmt_values = []
        for i, v in enumerate(vals):
            if not is_float_type[i] and leading_space:
                fmt_values.append(f" {_format(v)}")
            elif is_float_type[i]:
                fmt_values.append(_trim_zeros_single_float(float_format(v)))
            else:
                if leading_space is False:
                    # False specifically, so that the default is
                    # to include a space if we get here.
                    tpl = "{v}"
                else:
                    tpl = " {v}"
                fmt_values.append(tpl.format(v=_format(v)))

        return fmt_values


class DatarHTMLFormatter(HTMLFormatter):  # pragma: no cover
    """Fix nrows as we added one more row (dtype)"""

    def _write_regular_rows(
        self, fmt_values: Mapping[int, List[str]], indent: int
    ) -> None:
        is_truncated_horizontally = self.fmt.is_truncated_horizontally
        is_truncated_vertically = self.fmt.is_truncated_vertically

        nrows = len(self.fmt.tr_frame) + 1

        if self.fmt.index:
            fmtter = self.fmt._get_formatter("__index__")
            if fmtter is not None:
                index_values = self.fmt.tr_frame.index.map(fmtter)
            else:
                index_values = self.fmt.tr_frame.index.format()
        # dtype row
        index_values.insert(0, "")

        row: List[str] = []
        for i in range(nrows):

            if is_truncated_vertically and i == (self.fmt.tr_row_num):
                str_sep_row = ["..."] * len(row)
                tags = (
                    {
                        j: 'style="font-style: italic;" '
                        for j, _ in enumerate(str_sep_row)
                    }
                    if i == 0
                    else None
                )

                self.write_tr(
                    str_sep_row,
                    indent,
                    self.indent_delta,
                    tags=tags,
                    nindex_levels=self.row_levels,
                )

            row = []
            if self.fmt.index:
                row.append(index_values[i])
            # see gh-22579
            # Column misalignment also occurs for
            # a standard index when the columns index is named.
            # Add blank cell before data cells.
            elif self.show_col_idx_names:
                row.append("")
            row.extend(fmt_values[j][i] for j in range(self.ncols))

            if is_truncated_horizontally:
                dot_col_ix = self.fmt.tr_col_num + self.row_levels
                row.insert(dot_col_ix, "...")

            tags = (
                {j: 'style="font-style: italic;" ' for j, _ in enumerate(row)}
                if i == 0
                else None
            )

            self.write_tr(
                row,
                indent,
                self.indent_delta,
                tags=tags,
                nindex_levels=self.row_levels,
            )

    def _write_hierarchical_rows(
        self, fmt_values: Mapping[int, List[str]], indent: int
    ) -> None:
        template = 'rowspan="{span}" valign="top"'

        is_truncated_horizontally = self.fmt.is_truncated_horizontally
        is_truncated_vertically = self.fmt.is_truncated_vertically
        frame = self.fmt.tr_frame
        nrows = len(frame) + 1

        assert isinstance(frame.index, MultiIndex)
        idx_values = frame.index.format(
            sparsify=False, adjoin=False, names=False
        )
        # add dtype row
        len_idx_values = len(idx_values)
        idx_values = list(zip(*idx_values))
        idx_values.insert(0, ("",) * len_idx_values)

        if self.fmt.sparsify:
            # GH3547
            sentinel = lib.no_default
            levels = frame.index.format(
                sparsify=sentinel, adjoin=False, names=False
            )
            levels = [("",) + level for level in levels]
            level_lengths = get_level_lengths(levels, sentinel)
            inner_lvl = len(level_lengths) - 1
            if is_truncated_vertically:
                # Insert ... row and adjust idx_values and
                # level_lengths to take this into account.
                ins_row = self.fmt.tr_row_num
                inserted = False
                for lnum, records in enumerate(level_lengths):
                    rec_new = {}
                    for tag, span in list(records.items()):
                        if tag >= ins_row:
                            rec_new[tag + 1] = span
                        elif tag + span > ins_row:
                            rec_new[tag] = span + 1

                            # GH 14882 - Make sure insertion done once
                            if not inserted:
                                dot_row = list(idx_values[ins_row - 1])
                                dot_row[-1] = "..."
                                idx_values.insert(ins_row, tuple(dot_row))
                                inserted = True
                            else:
                                dot_row = list(idx_values[ins_row])
                                dot_row[inner_lvl - lnum] = "..."
                                idx_values[ins_row] = tuple(dot_row)
                        else:
                            rec_new[tag] = span
                        # If ins_row lies between tags, all cols idx cols
                        # receive ...
                        if tag + span == ins_row:
                            rec_new[ins_row] = 1
                            if lnum == 0:
                                idx_values.insert(
                                    ins_row, tuple(["..."] * len(level_lengths))
                                )

                            # GH 14882 - Place ... in correct level
                            elif inserted:
                                dot_row = list(idx_values[ins_row])
                                dot_row[inner_lvl - lnum] = "..."
                                idx_values[ins_row] = tuple(dot_row)
                    level_lengths[lnum] = rec_new

                level_lengths[inner_lvl][ins_row] = 1
                for ix_col in range(len(fmt_values)):
                    fmt_values[ix_col].insert(ins_row, "...")
                nrows += 1

            for i in range(nrows):
                row = []
                tags = {}

                sparse_offset = 0
                j = 0
                for records, v in zip(level_lengths, idx_values[i]):
                    if i in records:
                        if records[i] > 1:
                            tags[j] = (
                                template.format(span=records[i])
                                if i > 0
                                else 'style="font-style: italic;" '
                            )
                    else:
                        sparse_offset += 1
                        continue

                    j += 1
                    row.append(v)

                row.extend(fmt_values[j][i] for j in range(self.ncols))
                if i == 0:
                    tags.update(
                        {
                            j + k: 'style="font-style: italic;" '
                            for k in range(self.ncols)
                        }
                    )

                if is_truncated_horizontally:
                    row.insert(
                        self.row_levels - sparse_offset + self.fmt.tr_col_num,
                        "...",
                    )

                self.write_tr(
                    row,
                    indent,
                    self.indent_delta,
                    tags=tags,
                    nindex_levels=len(levels) - sparse_offset,
                )
        else:
            row = []
            for i in range(len(frame) + 1):
                if is_truncated_vertically and i == (self.fmt.tr_row_num):
                    str_sep_row = ["..."] * len(row)
                    tags = (
                        {
                            j: 'style="font-style: italic;" '
                            for j, _ in enumerate(str_sep_row)
                        }
                        if i == 0
                        else None
                    )

                    self.write_tr(
                        str_sep_row,
                        indent,
                        self.indent_delta,
                        tags=tags,
                        nindex_levels=self.row_levels,
                    )

                idx_values = list(
                    zip(
                        *frame.index.format(
                            sparsify=False, adjoin=False, names=False
                        )
                    )
                )
                idx_values.insert(0, ("",) * len_idx_values)
                row = []
                row.extend(idx_values[i])
                row.extend(fmt_values[j][i] for j in range(self.ncols))
                if is_truncated_horizontally:
                    row.insert(self.row_levels + self.fmt.tr_col_num, "...")

                tags = (
                    {
                        j: 'style="font-style: italic;" '
                        for j, _ in enumerate(row)
                    }
                    if i == 0
                    else None
                )

                self.write_tr(
                    row,
                    indent,
                    self.indent_delta,
                    tags=tags,
                    nindex_levels=frame.index.nlevels,
                )

    def render(self) -> List[str]:
        """Render the df"""
        from .grouped import DataFrameGroupBy, DataFrameRowwise

        self._write_table()

        if isinstance(self.frame, DataFrameRowwise):
            self.write(f"<p>Rowwise: {self.frame.attrs['_group_vars']}</p>")
        elif isinstance(self.frame, DataFrameGroupBy):
            ngroups = self.frame._group_data.shape[0]
            self.write(
                f"<p>Groups: {self.frame.attrs['_group_vars']} "
                f"(n={ngroups})</p>"
            )

        if self.should_show_dimensions:
            by = chr(215)  # ×
            self.write(
                f"<p>{len(self.frame)} rows {by} {len(self.frame.columns)} "
                "columns</p>"
            )

        return self.elements


class DatarNotebookFormatter(DatarHTMLFormatter, NotebookFormatter):
    """Notebook Formatter"""


class DatarStringFormatter(StringFormatter):  # pragma: no cover
    """String Formatter"""

    def to_string(self) -> str:
        """To string representation"""
        text = self._get_string_representation()
        grouping_info = getattr(self.fmt, "grouping_info")

        if grouping_info and self.fmt.should_show_dimensions:
            return "".join(
                [
                    # self.fmt.dimensions_info has two leading "\n"
                    text,
                    "\n",
                    grouping_info,
                    self.fmt.dimensions_info[1:],
                ]
            )
        if grouping_info:
            return "".join([text, "\n", grouping_info])
        if self.fmt.should_show_dimensions:
            return "".join([text, self.fmt.dimensions_info])

        return text


def _patch(option: bool) -> None:  # pragma: no cover
    """Patch pandas?"""
    if option:
        # monkey-patch the formatter
        fmt.DataFrameFormatter = DatarDataFrameFormatter
        fmt.GenericArrayFormatter = DatarGenericArrayFormatter
        html.HTMLFormatter = DatarHTMLFormatter
        html.NotebookFormatter = DatarNotebookFormatter
        stringf.StringFormatter = DatarStringFormatter
    else:
        fmt.DataFrameFormatter = DataFrameFormatter
        fmt.GenericArrayFormatter = GenericArrayFormatter
        html.HTMLFormatter = HTMLFormatter
        html.NotebookFormatter = NotebookFormatter
        stringf.StringFormatter = StringFormatter


add_option("frame_format_patch", True, _patch)



"""Specific verbs from this package"""
from typing import Any, List

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.types import is_scalar
from ..core.contexts import Context
from ..core.grouped import DataFrameGroupBy
from ..core.utils import get_option
from ..dplyr import select, slice_


@register_verb(DataFrame, context=Context.SELECT)
def get(
    _data: DataFrame,
    rows: Any = None,
    cols: Any = None,
    base0_: bool = None,
) -> Any:
    """Get a single element or a subset of a dataframe

    Args:
        _data: The dataframe
        rows: The rows to subset the dataframe
        cols: The columns to subset the dataframe
            If both rows and cols are scalar, then a single element will be
            returned
        base0_: If given by index, whether `rows` and `cols` are 0-based.
            If not given, will use `get_option("index.base.0")`

    Returns:
        A single element when both rows and cols are scalar, otherwise
        a subset of _data
    """
    base0_ = get_option("index.base.0", base0_)
    if isinstance(_data, DataFrameGroupBy):
        data = _data.copy(copy_grouped=True)
    else:
        data = _data.copy()
    # getting single element
    if (
        rows is not None
        and cols is not None
        and is_scalar(rows)
        and is_scalar(cols)
    ):
        if isinstance(rows, str):  # index
            rows = data.index.get_indexer_for([rows])[0]
        else:
            rows = rows - int(not base0_)
        if isinstance(cols, str):
            cols = data.columns.get_indexer_for([cols])[0]
        else:
            cols = cols - int(not base0_)
        return data.iloc[rows, cols]

    if cols is not None:
        data = select(
            data,
            cols,
            base0_=base0_,
            __calling_env=CallingEnvs.REGULAR,
        )

    if rows is not None:
        # slice only support integer index
        if not isinstance(rows, slice):
            if is_scalar(rows):
                rows = [rows]
            if not isinstance(rows[0], int):
                rows = data.index.get_indexer_for(rows) + int(not base0_)
        data = slice_(
            data,
            rows,
            base0_=base0_,
            __calling_env=CallingEnvs.REGULAR,
        )
    return data


@register_verb(DataFrame)
def flatten(_data: DataFrame, bycol: bool = False) -> List[Any]:
    """Flatten a dataframe into a 1-d python list

    Args:
        _data: The dataframe

    Returns:
        The flattened list
    """
    if bycol:
        return _data.T.values.flatten().tolist()
    return _data.values.flatten().tolist()


# @register_verb(DataFrame, context=Context.UNSET)
# def debug(
#         _data: DataFrameType,
#         *args: Any,
#         context: Union[Context, str] = Context.SELECT,
#         **kwargs: Any
# ) -> None: # not going any further
#     """Debug the expressions in the argument of the verb

#     Args:
#         _data: the dataframe
#         *args, **kwargs: The pipda expressions
#         context: How the expressions should be evaluated
#     """
#     def print_msg(msg: str, end: str = "\n"):
#         sys.stderr.write(f"[datar] DEBUG: {msg}{end}")

#     if isinstance(_data, DataFrameGroupBy):
#         print_msg("# DataFrameGroupBy:")
#         print_msg(_data.describe())
#     else:
#         print_msg("# DataFrame:")
#         print_msg(_data)

#     if args:
#         for i, arg in enumerate(args):
#             print_msg(f"# Arg#{i+1}")
#             print_msg("## Raw")
#             print_msg(arg)
#             print_msg("## Evaluated")
#             print_msg(evaluate_expr(arg, _data, context))

#     if kwargs:
#         for key, val in kwargs.items():
#             print_msg(f"# Kwarg#{key}")
#             print_msg("## Raw")
#             print_msg(val)
#             print_msg("## Evaluated")
#             print_msg(evaluate_expr(val, _data, context))


@register_verb(DataFrame)
def drop_index(_data: DataFrame) -> DataFrame:
    """Drop the index of a dataframe, works as a verb"""
    return _data.reset_index(drop=True)



"""Date time functions"""

import datetime
import functools
from typing import TYPE_CHECKING, Any, Iterable, List, Union

import numpy
import pandas
from pandas import DataFrame, Series
from pipda import register_func

from ..core.contexts import Context
from ..core.types import IntType, is_scalar, is_scalar_int
from .na import NA

if TYPE_CHECKING:
    from pandas import DatetimeIndex, Timestamp


@functools.singledispatch
def _as_date_dummy(
    x: Any,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[IntType, datetime.timedelta] = 0,
    origin: Any = None,
) -> datetime.date:
    """Convert a dummy object to date"""
    raise ValueError(f"Unable to convert to date with type: {type(x)!r}")


@_as_date_dummy.register(datetime.date)
def _(
    x: datetime.date,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[int, datetime.timedelta] = 0,
    origin: Any = None,
) -> datetime.date:
    if is_scalar_int(tz):
        tz = datetime.timedelta(hours=int(tz))

    return x + tz


@_as_date_dummy.register(datetime.datetime)
def _(
    x: datetime.datetime,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[IntType, datetime.timedelta] = 0,
    origin: Any = None,
):
    if is_scalar_int(tz):
        tz = datetime.timedelta(hours=int(tz))

    return (x + tz).date()


@_as_date_dummy.register(str)
def _(
    x: str,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[IntType, datetime.timedelta] = 0,
    origin: Any = None,
) -> datetime.date:
    if is_scalar_int(tz):
        tz = datetime.timedelta(hours=int(tz))

    try_formats = try_formats or [
        "%Y-%m-%d",
        "%Y/%m/%d",
        "%Y-%m-%d %H:%M:%S",
        "%Y/%m/%d %H:%M:%S",
    ]
    if not format:
        format = try_formats
    else:
        format = [format]

    for fmt in format:
        try:
            return (datetime.datetime.strptime(x, fmt) + tz).date()
        except ValueError:
            continue

    if optional:
        return NA

    raise ValueError("character string is not in a standard unambiguous format")


@_as_date_dummy.register(Series)
def _(
    x: Series,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[IntType, datetime.timedelta] = 0,
    origin: Any = None,
) -> datetime.date:
    return _as_date_dummy(
        x.values[0],
        format=format,
        try_formats=try_formats,
        optional=optional,
        tz=tz,
        origin=origin,
    )


@_as_date_dummy.register(int)
@_as_date_dummy.register(numpy.integer)
def _(
    x: IntType,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[IntType, datetime.timedelta] = 0,
    origin: Any = None,
) -> datetime.date:
    if isinstance(tz, (int, numpy.integer)):
        tz = datetime.timedelta(hours=int(tz))

    if isinstance(origin, str):
        origin = _as_date_dummy(origin)

    dt = origin + datetime.timedelta(days=int(x)) + tz

    if isinstance(dt, datetime.datetime):
        return dt.date()
    return dt


@register_func(None, context=Context.EVAL)
def as_date(
    x: DataFrame,
    format: str = None,
    try_formats: List[str] = None,
    optional: bool = False,
    tz: Union[IntType, datetime.timedelta] = 0,
    origin: Any = None,
) -> Union[Series, "Timestamp", "DatetimeIndex"]:
    """Convert an object to a datetime.date object

    See: https://rdrr.io/r/base/as.Date.html

    Args:
        x: Object that can be converted into a datetime.date object
        format:  If not specified, it will try try_formats one by one on
            the first non-NA element, and give an error if none works.
            Otherwise, the processing is via strptime
        try_formats: vector of format strings to try if format is not specified.
            Default formats to try:
            "%Y-%m-%d"
            "%Y/%m/%d"
            "%Y-%m-%d %H:%M:%S"
            "%Y/%m/%d %H:%M:%S"
        optional: indicating to return NA (instead of signalling an error)
            if the format guessing does not succeed.
        origin: a datetime.date/datetime object, or something which can be
            coerced by as_date(origin, ...) to such an object.
        tz: a time zone offset or a datetime.timedelta object.
            Note that time zone name is not supported yet.

    Returns:
        The datetime.date object
    """
    if not isinstance(x, Series):
        x = Series([x]) if is_scalar(x) else Series(x)

    out = x.transform(
        _as_date_dummy,
        format=format,
        try_formats=try_formats,
        optional=optional,
        tz=tz,
        origin=origin,
    )
    return pandas.to_datetime(out)


@register_func(None, context=Context.EVAL)
def as_pd_date(
    arg: Union[int, str, float, datetime.datetime, Iterable],
    *args: Any,
    **kwargs: Any,
) -> Union[Series, "Timestamp", "DatetimeIndex"]:
    """Alias of pandas.to_datetime(), but registered as a function
    so that it can be used in verbs.

    See https://pandas.pydata.org/docs/reference/api/pandas.to_datetime.html

    Args:
        arg: The argument to be converted to datetime
        *args: and
        **kwargs: Other arguments passing to `pandas.to_datetime()`

    Returns:
        Converted datetime
    """
    return pandas.to_datetime(arg, *args, **kwargs)



"""Relocate columns"""
from typing import Any, Union

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..base import setdiff, union
from .group_data import group_vars
from .select import _eval_select


@register_verb(DataFrame, context=Context.SELECT)
def relocate(
    _data: DataFrame,
    *args: Any,
    _before: Union[int, str] = None,
    _after: Union[int, str] = None,
    base0_: bool = None,
    **kwargs: Any,
) -> DataFrame:
    """change column positions

    See original API
    https://dplyr.tidyverse.org/reference/relocate.html

    Args:
        _data: A data frame
        *args: and
        **kwargs: Columns to rename and move
        _before: and
        _after: Destination. Supplying neither will move columns to
            the left-hand side; specifying both is an error.
        base0_: Whether `_before` and `_after` are 0-based if given by indexes.
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        An object of the same type as .data. The output has the following
        properties:
        - Rows are not affected.
        - The same columns appear in the output, but (usually) in a
            different place.
        - Data frame attributes are preserved.
        - Groups are not affected
    """
    gvars = group_vars(_data, __calling_env=CallingEnvs.REGULAR)
    all_columns = _data.columns
    to_move, new_names = _eval_select(
        all_columns,
        *args,
        **kwargs,
        base0_=base0_,
        _group_vars=gvars,
    )
    if _before is not None and _after is not None:
        raise ValueError("Must supply only one of `_before` and `_after`.")

    # length = len(all_columns)
    if _before is not None:
        where = min(
            _eval_select(
                all_columns,
                _before,
                _group_vars=[],
                base0_=base0_,
            )[0]
        )
        if where not in to_move:
            to_move.append(where)

    elif _after is not None:
        where = max(
            _eval_select(
                all_columns,
                _after,
                _group_vars=[],
                base0_=base0_,
            )[0]
        )
        if where not in to_move:
            to_move.insert(0, where)
    else:
        where = 0
        if where not in to_move:
            to_move.append(where)

    lhs = setdiff(range(where), to_move, __calling_env=CallingEnvs.REGULAR)
    rhs = setdiff(
        range(where + 1, _data.shape[1]),
        to_move,
        __calling_env=CallingEnvs.REGULAR,
    )
    pos = union(
        lhs,
        union(to_move, rhs, __calling_env=CallingEnvs.REGULAR),
        __calling_env=CallingEnvs.REGULAR,
    )
    out = _data.iloc[:, pos].copy()
    if new_names:
        out.rename(columns=new_names, inplace=True)

    return out



"""Extract a single colum

https://github.com/tidyverse/dplyr/blob/master/R/pull.R
"""

from typing import Mapping, Union

from pandas import DataFrame, Series
from pipda import register_verb

from ..core.contexts import Context
from ..core.utils import arg_match, df_getitem, position_at
from ..core.types import StringOrIter, ArrayLikeType, is_scalar


@register_verb(
    DataFrame, context=Context.SELECT, extra_contexts={"name": Context.EVAL}
)
def pull(
    _data: DataFrame,
    var: Union[int, str] = -1,
    name: StringOrIter = None,
    to: str = None,
    base0_: bool = None,
) -> Union[DataFrame, ArrayLikeType, Mapping[str, ArrayLikeType]]:

    """Pull a series or a dataframe from a dataframe

    Args:
        _data: The dataframe
        var: The column to pull, either the name or the index
        name: The name of the pulled value
            - If `to` is frame, or the value pulled is data frame, it will be
              the column names
            - If `to` is series, it will be the series name. If multiple names
              are given, only the first name will be used.
            - If `to` is series, but value pulled is a data frame, then a
              dictionary of series with the series names as keys or given `name`
              as keys.
        to: Type of data to return.
            Only works when pulling `a` for name `a$b`
            - series: Return a pandas Series object
              Group information will be lost
              If pulled value is a dataframe, it will return a dict of series,
              with the series names or the `name` provided.
            - array: Return a numpy.ndarray object
            - frame: Return a DataFrame with that column
            - list: Return a python list
            - dict: Return a dict with `name` as keys and pulled value as values
              Only a single column is allowed to pull
            - If not provided: `series` when pulled data has only one columns.
                `dict` if `name` provided and has the same length as the pulled
                single column. Otherwise `frame`.
        base0_: Whether `var` is 0-based if given by index
            If not provided, `datar.base.get_option('index.base.0')` is used.

    Returns:
        The data according to `to`
    """
    # make sure pull(df, 'x') pulls a dataframe for columns
    # x$a, x$b in df

    to = arg_match(to, "to", ["list", "array", "frame", "series", "dict", None])
    if name is not None and is_scalar(name):
        name = [name]  # type: ignore

    if isinstance(var, int):
        var = position_at(var, _data.shape[1], base0=base0_)
        var = _data.columns[var]
        var = var.split("$", 1)[0]

    pulled = df_getitem(_data, var)
    # if var in _data.columns and isinstance(pulled, DataFrame):
    #     pulled = pulled.iloc[:, 0]

    if to is None:
        if name is not None and len(name) == len(pulled):
            to = "dict"
        else:
            to = "frame" if isinstance(pulled, DataFrame) else "series"

    if to == "dict":
        if name is None or len(name) != len(pulled):
            raise ValueError(
                "No `name` provided or length mismatches with the values."
            )
        return dict(zip(name, pulled))
    if to == "list":
        return pulled.values.tolist()
    if to == "array":
        return pulled.values
    if to == "frame":
        value = pulled if isinstance(pulled, DataFrame) else pulled.to_frame()
        if name and len(name) != value.shape[1]:
            raise ValueError(
                f"Expect {value.shape[1]} names but got {len(name)}."
            )
        if name:
            value.columns = name
        return value
    # if to == 'series':
    if isinstance(pulled, DataFrame) and pulled.shape[1] == 1:
        pulled = pulled.iloc[:, -1]
    if isinstance(pulled, Series):
        if name:
            pulled.name = name[0]
        return pulled
    # df
    if name and len(name) != pulled.shape[1]:
        raise ValueError(f"Expect {pulled.shape[1]} names but got {len(name)}.")

    out = pulled.to_dict("series")
    if not name:
        return out

    for newname, oldname in zip(name, out):
        out[newname] = out.pop(oldname)
    return out



"""Replace NAs with specified values"""
from functools import singledispatch
from typing import Any, Iterable

import numpy
import pandas
from pandas import DataFrame
from pandas.core.series import Series
from pipda import register_verb

from ..core.types import ArrayLikeType, is_scalar
from ..core.contexts import Context


@singledispatch
def _replace_na(data: Iterable[Any], replace: Any) -> Iterable[Any]:
    """Replace NA for any iterables"""
    return type(data)(  # type: ignore
        replace if is_scalar(elem) and pandas.isnull(elem) else elem
        for elem in data
    )


@_replace_na.register(numpy.ndarray)
@_replace_na.register(Series)
def _(data: ArrayLikeType, replace: Any) -> ArrayLikeType:
    """Replace NA for numpy.ndarray or Series"""
    ret = data.copy()
    ret[pandas.isnull(ret)] = replace
    return ret


@_replace_na.register(DataFrame)
def _(data: DataFrame, replace: Any) -> DataFrame:
    """Replace NA for numpy.ndarray or DataFrame"""
    # TODO: allow replace to be a list as an entire value to replace
    return data.fillna(replace)


@register_verb(
    (DataFrame, Series, numpy.ndarray, list, tuple, set), context=Context.EVAL
)
def replace_na(
    data: Iterable[Any],
    data_or_replace: Any = None,
    replace: Any = None,
) -> Any:
    """Replace NA with a value

    This function can be also used not as a verb. As a function called as
    an argument in a verb, data is passed implicitly. Then one could
    pass data_or_replace as the data to replace.

    Args:
        data: The data piped in
        data_or_replace: When called as argument of a verb, this is the
            data to replace. Otherwise this is the replacement.
        replace: The value to replace with
            Can only be a scalar or dict for data frame.
            So replace NA with a list is not supported yet.

    Returns:
        Corresponding data with NAs replaced
    """
    if data_or_replace is None and replace is None:
        return data.copy()

    if replace is None:
        # no replace, then data_or_replace should be replace
        replace = data_or_replace
    else:
        # replace specified, determine data
        # If data_or_replace is specified, it's data
        data = data if data_or_replace is None else data_or_replace

    return _replace_na(data, replace)



"""Let datar warn when builtin names are tried to be imported"""
import sys
from typing import Callable, Any
from executing import Source

WARNED = set()


def warn_builtin_names(**names: Callable) -> Callable[[str], Any]:
    """Generate __getattr__ function to warn the builtin names"""
    from .utils import logger, get_option

    # Enables tempoarory warn on or off
    warn = True

    def _getattr(name: str):
        """A route to let us check if the function is imported by
            >>> from module import func
        But not
            >>> import module
            >>> module.func

        If func is a python built-in function, then it gets overriden by datar.

        The side effects:
        1. This executes every time when `module.func` is called
        2. When the source is not avaiable for `module.func`, it will be a
           false alarm treat this as `from module import func`
        3. It warns even you do `from module import func as alias`

        Instead, you can do `from module import func_` to access function `func`

        Args:
            **names: The name-module pairs

        Returns:
            A function that can be used as `__getattr__` for a module
        """
        nonlocal warn
        if name == "_warn":
            warn = True
            return None
        if name == "_no_warn":
            warn = False
            return None

        if name == "__path__" or name not in names:
            raise AttributeError

        if warn and name not in WARNED and get_option("warn.builtin.names"):
            node = Source.executing(sys._getframe(1)).node
            if not node:
                WARNED.add(name)
                logger.warning(
                    'Builtin name "%s" has been overriden by datar.', name
                )
        return names[name]

    return _getattr



"""Grouping metadata"""

from typing import List
from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.grouped import DataFrameGroupBy
from ..base import setdiff


@register_verb(DataFrame)
def group_data(_data: DataFrame) -> DataFrame:
    """Returns a data frame that defines the grouping structure.

    Args:
        _data: The data frame

    Returns:
        The columns give the values of the grouping variables. The last column,
        always called `_rows`, is a list of integer vectors that gives the
        location of the rows in each group.

        Note that `_rows` are always 0-based.
    """
    rows = list(range(_data.shape[0]))
    return DataFrame({"_rows": [rows]})


@group_data.register(DataFrameGroupBy)
def _(_data: DataFrameGroupBy) -> DataFrame:
    return _data._group_data


@register_verb(DataFrame)
def group_keys(_data: DataFrame) -> DataFrame:
    """Just grouping data without the `_rows` columns

    Note:
        Additional arguments to select columns in dplyr are deprecated.
        Here we don't support it ahead.

    Args:
        _data: The data frame

    Returns:
        The group data without `_rows` column.
    """
    return DataFrame(index=[0])


@group_keys.register(DataFrameGroupBy)
def _(_data: DataFrameGroupBy) -> DataFrame:
    # .copy() allows future modifications
    return (
        group_data(_data, __calling_env=CallingEnvs.REGULAR).iloc[:, :-1].copy()
    )


@register_verb(DataFrame)
def group_rows(_data: DataFrame) -> List[List[int]]:
    """The locations of grouping structure, always 0-based."""
    rows = list(range(_data.shape[0]))
    return [rows]


@group_rows.register(DataFrameGroupBy)
def _(_data: DataFrame) -> List[List[int]]:
    return group_data(_data, __calling_env=CallingEnvs.REGULAR)[
        "_rows"
    ].tolist()


@register_verb(DataFrame)
def group_indices(_data: DataFrame) -> List[int]:
    """Returns an integer vector the same length as `_data`.

    Always 0-based.
    """
    return [0] * _data.shape[0]


@group_indices.register(DataFrameGroupBy)
def _(_data: DataFrameGroupBy) -> List[int]:
    ret = {}
    for row in group_data(
        _data, __calling_env=CallingEnvs.REGULAR
    ).itertuples():
        for index in row[-1]:
            ret[index] = row.Index
    return [ret[key] for key in sorted(ret)]


@register_verb(DataFrame)
def group_vars(_data: DataFrame) -> List[str]:
    """Gives names of grouping variables as character vector"""
    # If it is a subdf of a DataFrameGroupBy, still be able to
    # return the group_vars
    index = _data.attrs.get("_group_index", None)
    if index is None:
        return []
    gdata = _data.attrs["_group_data"]
    return setdiff(gdata.columns, ["_rows"], __calling_env=CallingEnvs.REGULAR)


@group_vars.register(DataFrameGroupBy)
def _(_data: DataFrameGroupBy) -> List[str]:
    return _data.attrs["_group_vars"]


# groups in dplyr returns R list
groups = group_vars
group_cols = group_vars


@register_verb(DataFrame)
def group_size(_data: DataFrame) -> List[int]:
    """Gives the size of each group"""
    return [_data.shape[0]]


@group_size.register(DataFrameGroupBy)
def _(_data: DataFrameGroupBy) -> List[int]:
    return list(
        map(len, group_data(_data, __calling_env=CallingEnvs.REGULAR)["_rows"])
    )


@register_verb(DataFrame)
def n_groups(_data: DataFrame) -> int:
    """Gives the total number of groups."""
    return 1


@n_groups.register(DataFrameGroupBy)
def _(_data: DataFrameGroupBy) -> int:
    return group_data(_data, __calling_env=CallingEnvs.REGULAR).shape[0]



"""Pack and unpack

https://github.com/tidyverse/tidyr/blob/master/R/pack.R
"""
from typing import Iterable, Set, Union, Callable

from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.utils import vars_select, copy_attrs, reconstruct_tibble
from ..core.contexts import Context
from ..core.types import StringOrIter, IntOrIter, is_scalar
from ..core.names import repair_names
from ..core.grouped import DataFrameGroupBy

from ..base import setdiff
from ..dplyr import bind_cols


@register_verb(DataFrame, context=Context.SELECT)
def pack(
    _data: DataFrame,
    _names_sep: str = None,
    base0_: bool = None,
    **cols: Union[str, int],
) -> DataFrame:
    """Makes df narrow by collapsing a set of columns into a single df-column.

    Args:
        _data: A data frame
        **cols: Columns to pack
        _names_sep: If `None`, the default, the names will be left as is.
            Inner names will come from the former outer names
            If a string, the inner and outer names will be used together.
            The names of the new outer columns will be formed by pasting
            together the outer and the inner column names, separated by
            `_names_sep`.
        base0_: Whether `**cols` are 0-based
            if not provided, will use `datar.base.get_option('index.base.0')`
    """
    if not cols:
        return _data.copy()

    from .nest import _strip_names

    all_columns = _data.columns
    colgroups = {}
    usedcols = set()
    for group, columns in cols.items():
        oldcols = all_columns[vars_select(all_columns, columns, base0=base0_)]
        usedcols = usedcols.union(oldcols)
        newcols = (
            oldcols
            if _names_sep is None
            else _strip_names(oldcols, group, _names_sep)
        )
        colgroups[group] = zip(newcols, oldcols)

    cols = {}
    for group, columns in colgroups.items():
        for newcol, oldcol in columns:
            cols[f"{group}${newcol}"] = _data[oldcol]

    asis = setdiff(
        _data.columns,
        usedcols,
        __calling_env=CallingEnvs.REGULAR,
    )
    out = bind_cols(
        _data[asis],
        DataFrame(cols),
        __calling_env=CallingEnvs.REGULAR,
    )
    return reconstruct_tibble(_data, out, keep_rowwise=True)


@register_verb(DataFrame, context=Context.SELECT)
def unpack(
    data: DataFrame,
    cols: Union[StringOrIter, IntOrIter],
    names_sep: str = None,
    names_repair: Union[str, Callable] = "check_unique",
    base0_: bool = None,
) -> DataFrame:
    """Makes df wider by expanding df-columns back out into individual columns.

    For empty columns, the column is kept asis, instead of removing it.

    Args:
        data: A data frame
        cols: Columns to unpack
        names_sep: If `None`, the default, the names will be left as is.
            Inner names will come from the former outer names
            If a string, the inner and outer names will be used together.
            The names of the new outer columns will be formed by pasting
            together the outer and the inner column names, separated by
            `_names_sep`.
        name_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether `cols` are 0-based
            if not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Data frame with given columns unpacked.
    """
    if is_scalar(cols):
        cols = [cols]

    all_columns = data.columns
    cols = _check_present(
        data,
        cols,
        all_columns,
        base0=base0_,
    )

    out = (
        data.copy(copy_grouped=True)
        if isinstance(data, DataFrameGroupBy)
        else data.copy()
    )
    new_cols = []
    for col in data.columns:
        if "$" in col:
            parts = col.split("$", 1)
            if parts[0] not in cols:
                new_cols.append(col)
            else:
                replace = "" if names_sep is None else f"{parts[0]}{names_sep}"
                new_cols.append(f"{replace}{parts[1]}")
        # elif col in cols: # empty list column
        #     # remove it from out
        #     out.drop(columns=col, inplace=True)
        else:
            new_cols.append(col)

    new_cols = repair_names(new_cols, names_repair, base0_)
    out.columns = new_cols

    copy_attrs(out, data)
    return out


def _check_present(
    data: DataFrame,
    cols: Iterable[Union[int, str]],
    all_columns: Iterable[str],
    base0: bool = None,
) -> Set[str]:
    """Check if cols are packed columns"""
    out = set()
    for col in cols:
        if not isinstance(col, str):
            columns = vars_select(all_columns, col, base0=base0)
            columns = all_columns[columns][0].split("$", 1)[0]
        else:
            columns = [col]

        for column in columns:
            if not (column in data and len(data[column]) == 0) and not any(
                allcol.startswith(f"{column}$") for allcol in all_columns
            ):
                raise ValueError(f"`{column}` must be a data frame column.")

            if column in out:
                raise ValueError(
                    f"`{column}` has already been selected. "
                    "Number of packed columns also counts when "
                    "selecting using indexes."
                )
            out.add(column)
    return out



"""Pivot data from wide to long

https://github.com/tidyverse/tidyr/blob/HEAD/R/pivot-long.R
"""
import re
from typing import Iterable, Mapping, Callable, Union

import pandas
from pandas import DataFrame
from pandas.core.dtypes.common import is_categorical_dtype
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.defaults import DEFAULT_COLUMN_PREFIX
from ..core.contexts import Context
from ..core.types import NAType, StringOrIter, Dtype, is_scalar
from ..core.utils import vars_select, apply_dtypes, reconstruct_tibble
from ..core.names import repair_names

from ..base import setdiff, union
from ..dplyr import relocate

from .extract import extract
from .separate import separate


@register_verb(DataFrame, context=Context.SELECT)
def pivot_longer(
    _data: DataFrame,
    cols: StringOrIter,
    names_to: Union[NAType, str, Iterable[Union[NAType, str]]] = "name",
    names_prefix: str = None,
    names_sep: str = None,
    names_pattern: str = None,
    names_ptypes: Union[Dtype, Mapping[str, Dtype]] = None,
    names_transform: Union[Callable, Mapping[str, Callable]] = None,
    names_repair="check_unique",
    values_to: str = "value",
    values_drop_na: bool = False,
    values_ptypes: Union[Dtype, Mapping[str, Dtype]] = None,
    values_transform: Union[Callable, Mapping[str, Callable]] = None,
    base0_: bool = None,
):
    """ "lengthens" data, increasing the number of rows and
    decreasing the number of columns.

    The row order is a bit different from `tidyr` and `pandas.DataFrame.melt`.
        >>> df = tibble(x=f[1:2], y=f[3:4])
        >>> pivot_longer(df, f[f.x:f.y])
        >>> #    name   value
        >>> # 0  x      1
        >>> # 1  x      2
        >>> # 2  y      3
        >>> # 3  y      4
    But with `tidyr::pivot_longer`, the output will be:
        >>> # # A tibble: 4 x 2
        >>> # name  value
        >>> # <chr> <int>
        >>> # 1 x   1
        >>> # 2 y   3
        >>> # 3 x   2
        >>> # 4 y   4

    Args:
        _data: A data frame to pivot.
        cols: Columns to pivot into longer format.
        names_to: A string specifying the name of the column to create from
            the data stored in the column names of data.
            Can be a character vector, creating multiple columns, if names_sep
            or names_pattern is provided. In this case, there are two special
            values you can take advantage of:
            - `None`/`NA`/`NULL` will discard that component of the name.
            - `.value`/`_value` indicates that component of the name defines
                the name of the column containing the cell values,
                overriding values_to.
            - Different as `tidyr`: With `.value`/`_value`, if there are other
              parts of the names to distinguish the groups, they must be
              captured. For example, use `r'(\\w)_(\\d)'` to match `'a_1'` and
              `['.value', NA]` to discard the suffix, instead of use
              `r'(\\w)_\\d'` to match.
        names_prefix: A regular expression used to remove matching text from
            the start of each variable name.
        names_sep: and
        names_pattern: If names_to contains multiple values,
            these arguments control how the column name is broken up.
            names_sep takes the same specification as separate(), and
            can either be a numeric vector (specifying positions to break on),
            or a single string (specifying a regular expression to split on).
        names_pattern: takes the same specification as extract(),
            a regular expression containing matching groups (()).
        names_ptypes: and
        values_ptypes: A list of column name-prototype pairs.
            A prototype (or ptype for short) is a zero-length vector
            (like integer() or numeric()) that defines the type, class, and
            attributes of a vector. Use these arguments if you want to confirm
            that the created columns are the types that you expect.
            Note that if you want to change (instead of confirm) the types
            of specific columns, you should use names_transform or
            values_transform instead.
        names_transform: and
        values_transform: A list of column name-function pairs.
            Use these arguments if you need to change the types of
            specific columns. For example,
            names_transform = dict(week = as.integer) would convert a
            character variable called week to an integer.
            If not specified, the type of the columns generated from names_to
            will be character, and the type of the variables generated from
            values_to will be the common type of the input columns used to
            generate them.
        names_repair: Not supported yet.
        values_to: A string specifying the name of the column to create from
            the data stored in cell values. If names_to is a character
            containing the special `.value`/`_value` sentinel, this value
            will be ignored, and the name of the value column will be derived
            from part of the existing column names.
        values_drop_na: If TRUE, will drop rows that contain only NAs in
            the value_to column. This effectively converts explicit missing
            values to implicit missing values, and should generally be used
            only when missing values in data were created by its structure.
        names_repair: treatment of problematic column names:
            - "minimal": No name repair or checks, beyond basic existence,
            - "unique": Make sure names are unique and not empty,
            - "check_unique": (default value), no name repair,
                but check they are unique,
            - "universal": Make the names unique and syntactic
            - a function: apply custom name repair
        base0_: Whether `cols` are 0-based if given by indexes
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        The pivoted dataframe.
    """
    rowid_column = "_PIVOT_ROWID_"
    ret = _data.assign(**{rowid_column: range(_data.shape[0])})
    all_columns = ret.columns
    columns = _data.columns[vars_select(_data.columns, cols, base0=base0_)]
    id_columns = all_columns.difference(columns)

    if is_scalar(names_to):
        names_to = [names_to]  # type: ignore

    tmp_names_to = []
    # We need to NA/names to be kept for .value pivot
    na_names_to = []
    for i, name in enumerate(names_to):
        if pandas.isnull(name):
            na_name = f"__{DEFAULT_COLUMN_PREFIX}_NA_{i}__"
            na_names_to.append(na_name)
            tmp_names_to.append(na_name)
        elif name == "_value":
            tmp_names_to.append(".value")
        else:
            tmp_names_to.append(name)
    names_to = tmp_names_to

    if len(names_to) > 1 and not names_sep and not names_pattern:
        raise ValueError(
            "If you supply multiple names in `names_to` you must also "
            "supply one of `names_sep` or `names_pattern`."
        )

    if names_sep and names_pattern:
        raise ValueError(
            "Only one of `names_sep` or `names_pattern` should be supplied."
        )

    var_name = "__tmp_names_to__" if names_pattern or names_sep else names_to[0]
    ret = ret.melt(
        id_vars=id_columns,
        # Use the rest columns automatically.
        # Don't specify so that duplicated column names can be used.
        # value_vars=columns,
        var_name=var_name,
        value_name=values_to,
    )
    if names_prefix:
        names_prefix = re.compile(f"^{re.escape(names_prefix)}")
        ret[var_name] = ret[var_name].str.replace(names_prefix, "")

    if all(is_categorical_dtype(_data[col]) for col in columns):
        ret[values_to] = ret[values_to].astype("category")

    if names_pattern:
        ret = extract(
            ret,
            var_name,
            into=names_to,
            regex=names_pattern,
            __calling_env=CallingEnvs.REGULAR,
        )

    if names_sep:
        ret = separate(
            ret,
            var_name,
            into=names_to,
            sep=names_sep,
            __calling_env=CallingEnvs.REGULAR,
        )
    # extract/separate puts `into` last
    ret = relocate(
        ret,
        values_to,
        _after=-1,
        base0_=True,
        __calling_env=CallingEnvs.REGULAR,
    )

    if ".value" in names_to:
        names_to = setdiff(
            names_to,
            [".value"],
            __calling_env=CallingEnvs.REGULAR,
        )
        index_columns = union(
            id_columns,
            names_to,
            __calling_env=CallingEnvs.REGULAR,
        )
        names_to = setdiff(
            names_to,
            na_names_to,
            __calling_env=CallingEnvs.REGULAR,
        )

        # keep the order
        value_columns = pandas.unique(ret[".value"].values)
        ret.set_index(index_columns, inplace=True)
        ret.index = list(ret.index)
        ret2 = ret.pivot(columns=".value", values=values_to).reset_index()
        id_data = DataFrame(ret2["index"].tolist(), columns=index_columns)
        ret = pandas.concat(
            [
                id_data[
                    id_data.columns.difference(na_names_to).difference(
                        [rowid_column]
                    )
                ],
                ret2[value_columns],
            ],
            axis=1,
        )
        values_to = value_columns
    else:
        values_to = [values_to]
        ret.drop(columns=[rowid_column], inplace=True)

    if values_drop_na:
        ret.dropna(subset=values_to, inplace=True)

    names_data = ret.loc[:, names_to]  # SettingwithCopyWarning
    apply_dtypes(names_data, names_ptypes)
    ret[names_to] = names_data

    values_data = ret[values_to]
    apply_dtypes(values_data, values_ptypes)
    ret[values_to] = values_data

    if names_transform:
        for name in names_to:
            if callable(names_transform):
                ret[name] = ret[name].apply(names_transform)
            elif name in names_transform:
                ret[name] = ret[name].apply(names_transform[name])

    if values_transform:
        for name in values_to:
            if callable(values_transform):
                ret[name] = ret[name].apply(values_transform)
            elif name in values_transform:
                ret[name] = ret[name].apply(values_transform[name])

    names = repair_names(ret.columns.tolist(), names_repair, base0_=base0_)
    ret.columns = names

    return reconstruct_tibble(_data, ret)



"""Functions from tidyr"""

from typing import Iterable

from pipda import register_func

from ..core.types import NumericType
from ..core.contexts import Context
from ..base import seq


@register_func(None, context=Context.EVAL)
def full_seq(
    x: Iterable[NumericType], period: NumericType, tol: float = 1e-6
) -> Iterable[NumericType]:
    """Create the full sequence of values in a vector

    Args:
        x: A numeric vector.
        period: Gap between each observation. The existing data will be
            checked to ensure that it is actually of this periodicity.
        tol: Numerical tolerance for checking periodicity.

    Returns:
        The full sequence
    """
    minx = min(x)  # na not counted
    maxx = max(x)

    if any(
        (elem - minx) % period > tol and period - ((elem - minx) % period) > tol
        for elem in x
    ):
        raise ValueError("`x` is not a regular sequence.")

    if period - ((maxx - minx) % period) <= tol:
        maxx += tol

    return seq(minx, maxx, by=period)



"""Provides functions to add or remove levels"""
from typing import Any, Iterable, List

from pandas import Categorical
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..base import levels, union, table, intersect, setdiff
from ..core.contexts import Context
from ..core.types import ForcatsRegType, ForcatsType, is_scalar, is_null
from .lvls import lvls_expand, lvls_union, refactor
from .utils import check_factor


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_expand(_f: ForcatsType, *additional_levels: Any) -> Categorical:
    """Add additional levels to a factor

    Args:
        _f: A factor
        *additional_levels: Additional levels to add to the factor.
            Levels that already exist will be silently ignored.

    Returns:
        The factor with levels expanded
    """
    _f = check_factor(_f)
    levs = levels(_f, __calling_env=CallingEnvs.REGULAR)
    addlevs = []
    for alev in additional_levels:
        if is_scalar(alev):
            addlevs.append(alev)
        else:
            addlevs.extend(alev)
    new_levels = union(levs, addlevs)
    return lvls_expand(_f, new_levels, __calling_env=CallingEnvs.REGULAR)


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_explicit_na(
    _f: ForcatsType, na_level: Any = "(Missing)"
) -> Categorical:
    """Make missing values explicit

    This gives missing values an explicit factor level, ensuring that they
    appear in summaries and on plots.

    Args:
        _f: A factor
        na_level: Level to use for missing values.
            This is what NAs will be changed to.

    Returns:
        The factor with explict na_levels
    """
    _f = check_factor(_f)
    # levs = levels(_f, __calling_env=CallingEnvs.REGULAR)
    is_missing = is_null(_f)
    # is_missing_level = is_null(levs)

    if any(is_missing):
        _f = fct_expand(_f, na_level)
        _f[is_missing] = na_level
        return _f

    # NAs cannot be a level in pandas.Categorical
    # if any(is_missing_level):
    #     levs[is_missing_level] = na_level
    #     return lvls_revalue(_f, levs)

    return _f


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_drop(_f: ForcatsType, only: Any = None) -> Categorical:
    """Drop unused levels

    Args:
        _f: A factor
        only: A character vector restricting the set of levels to be dropped.
            If supplied, only levels that have no entries and appear in
            this vector will be removed.

    Returns:
        The factor with unused levels dropped
    """
    _f = check_factor(_f)

    levs = levels(_f, __calling_env=CallingEnvs.REGULAR)
    count = table(_f, __calling_env=CallingEnvs.REGULAR).iloc[0, :]

    to_drop = levs[count == 0]
    if only is not None and is_scalar(only):
        only = [only]

    if only is not None:
        to_drop = intersect(to_drop, only, __calling_env=CallingEnvs.REGULAR)

    return refactor(
        _f,
        new_levels=setdiff(levs, to_drop, __calling_env=CallingEnvs.REGULAR),
    )


@register_verb(ForcatsRegType, context=Context.EVAL)
def fct_unify(
    fs: Iterable[ForcatsType],
    levels: Iterable = None,
) -> List[Categorical]:
    """Unify the levels in a list of factors

    Args:
        fs: A list of factors
        levels: Set of levels to apply to every factor. Default to union
            of all factor levels

    Returns:
        A list of factors with the levels expanded
    """
    if levels is None:
        levels = lvls_union(fs)

    out = []
    for fct in fs:
        fct = check_factor(fct)
        out.append(
            lvls_expand(
                fct,
                new_levels=levels,
                __calling_env=CallingEnvs.REGULAR,
            )
        )
    return out



"""Separate a character column into multiple columns with a regular
expression or numeric locations

https://github.com/tidyverse/tidyr/blob/HEAD/R/separate.R
"""
import re
from typing import Any, Iterable, List, Tuple, Union, Mapping

import pandas
from pandas import DataFrame
from pipda import register_verb
from pipda.utils import CallingEnvs

from ..core.contexts import Context
from ..core.types import Dtype, NAType, is_scalar
from ..core.grouped import DataFrameGroupBy
from ..core.utils import (
    logger,
    vars_select,
    apply_dtypes,
    position_at,
    reconstruct_tibble,
)

from ..base import NA
from ..dplyr import ungroup, mutate

from .chop import unchop


@register_verb(DataFrame, context=Context.SELECT)
def separate(
    data: DataFrame,
    col: Union[str, int],
    into: Union[NAType, str, Iterable[Union[NAType, str]]],
    sep: Union[int, str] = r"[^0-9A-Za-z]+",
    remove: bool = True,
    convert: Union[bool, Dtype, Mapping[str, Dtype]] = False,
    extra: str = "warn",
    fill: str = "warn",
    base0_: bool = None,
) -> DataFrame:
    """Given either a regular expression or a vector of character positions,
    turns a single character column into multiple columns.

    Args:
        data: The dataframe
        col: Column name or position.
        into: Names of new variables to create as character vector.
            Use `None`/`NA`/`NULL` to omit the variable in the output.
        sep: Separator between columns.
            If str, `sep` is interpreted as a regular expression.
            The default value is a regular expression that matches
            any sequence of non-alphanumeric values.
            If int, `sep` is interpreted as character positions to split at.
        remove: If TRUE, remove input column from output data frame.
        convert: The universal type for the extracted columns or a dict for
            individual ones
            Note that when given `TRUE`, `DataFrame.convert_dtypes()` is called,
            but it will not convert `str` to other types
            (For example, `'1'` to `1`). You have to specify the dtype yourself.
        extra: If sep is a character vector, this controls what happens when
            there are too many pieces. There are three valid options:
            - "warn" (the default): emit a warning and drop extra values.
            - "drop": drop any extra values without a warning.
            - "merge": only splits at most length(into) times
        fill: If sep is a character vector, this controls what happens when
            there are not enough pieces. There are three valid options:
            - "warn" (the default): emit a warning and fill from the right
            - "right": fill with missing values on the right
            - "left": fill with missing values on the left
        base0_: Whether `col` is 0-based when given by index and Whether `sep`
            is 0-based if given by position
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Dataframe with separated columns.
    """
    if is_scalar(into):
        into = [into]  # type: ignore

    if not all(isinstance(it, str) or pandas.isnull(it) for it in into):
        raise ValueError("`into` must be a string or a list of strings.")

    all_columns = data.columns
    col = vars_select(all_columns, col, base0=base0_)
    col = all_columns[col[0]]

    colindex = [i for i, outcol in enumerate(into) if not pandas.isnull(outcol)]
    non_na_elems = lambda row: [row[i] for i in colindex]
    # series.str.split can't do extra and fill
    # extracted = data[col].str.split(sep, expand=True).iloc[:, colindex]
    nout = len(into)
    extra_warns = []
    missing_warns = []

    separated = data[col].apply(
        _separate_col,
        nout=nout,
        sep=sep,
        extra=extra,
        fill=fill,
        base0=base0_,
        extra_warns=extra_warns,
        missing_warns=missing_warns,
    )

    if extra_warns:
        logger.warning(
            "Expected %s pieces. " "Additional pieces discarded in %s rows %s.",
            nout,
            len(extra_warns),
            extra_warns,
        )
    if missing_warns:
        logger.warning(
            "Expected %s pieces. "
            "Missing pieces filled with `NA` in %s rows %s.",
            nout,
            len(missing_warns),
            missing_warns,
        )

    separated = DataFrame(separated.values.tolist()).iloc[:, colindex]
    separated.columns = non_na_elems(into)
    apply_dtypes(separated, convert)

    out = data.drop(columns=[col]) if remove else data
    out = mutate(out, separated, __calling_env=CallingEnvs.REGULAR)

    return reconstruct_tibble(data, out)


@register_verb(DataFrame, context=Context.SELECT)
def separate_rows(
    data: DataFrame,
    *columns: Tuple[str],
    sep: str = r"[^0-9A-Za-z]+",
    convert: Union[bool, Dtype, Mapping[str, Dtype]] = False,
    base0_: bool = None,
) -> DataFrame:
    """Separates the values and places each one in its own row.

    Args:
        data: The dataframe
        *columns: The columns to separate on
        sep: Separator between columns.
        convert: The universal type for the extracted columns or a dict for
            individual ones
        base0_: Whether `columns` is 0-based when given by index
            If not provided, will use `datar.base.get_option('index.base.0')`

    Returns:
        Dataframe with rows separated and repeated.
    """
    all_columns = data.columns
    selected = all_columns[vars_select(all_columns, *columns, base0=base0_)]
    out = (
        data.copy(copy_grouped=True)
        if isinstance(data, DataFrameGroupBy)
        else data.copy()
    )
    for sel in selected:
        out[sel] = out[sel].apply(
            _separate_col,
            nout=0,
            sep=sep,
            extra="merge",
            fill="right",
            base0=base0_,
            extra_warns=[],
            missing_warns=[],
        )

    out = unchop(
        out,
        selected,
        keep_empty=True,
        ptype=convert,
        base0_=base0_,
        __calling_env=CallingEnvs.REGULAR,
    )
    return reconstruct_tibble(
        out,
        ungroup(out, __calling_env=CallingEnvs.REGULAR),
        selected,
        keep_rowwise=True,
    )


def _separate_col(
    elem: Any,
    nout: int,
    sep: Union[str, int],
    extra: str,
    fill: str,
    base0: bool,

    extra_warns: List[str] = [],  # mutatable to save warnings
    missing_warns: List[str] = [],
) -> List[Union[str, NAType]]:
    """Separate the column"""
    if (is_scalar(elem) and pandas.isnull(elem)) or (
        not is_scalar(elem) and any(pandas.isnull(elem))
    ):
        return [NA] * nout if nout > 0 else NA  # type: ignore

    elem = str(elem)
    if isinstance(sep, int):
        try:
            tmp = position_at(sep, len(elem), base0)
        except IndexError:
            tmp = 0 if sep < 0 else len(elem) - 1
        tmp = sep - 1 if sep < 0 else tmp
        row = [elem[: tmp + 1], elem[tmp + 1 :]]
    else:
        row = re.split(sep, elem, 0 if nout == 0 else nout - 1)
    if nout == 0:
        return row
    if len(row) < nout:
        if fill == "warn" and (
            not missing_warns or missing_warns[-1] != "...truncated"
        ):
            missing_warns.append(elem)
        if fill in ("warn", "right"):
            row += [NA] * (nout - len(row))  # type: ignore
        else:
            row = [NA] * (nout - len(row)) + row
    elif not isinstance(sep, int):
        more_splits = re.split(sep, row[-1], 1)
        if len(more_splits) > 1:
            if extra == "warn" and (
                not extra_warns or extra_warns[-1] != "...truncated"
            ):
                extra_warns.append(elem)
            if extra in ("warn", "drop"):
                row[-1] = more_splits[0]
    return row



"""Some constants/alias for R counterparts"""
import math
from string import ascii_letters

import numpy


pi = math.pi

letters = numpy.array(list(ascii_letters[:26]))
LETTERS = numpy.array(list(ascii_letters[26:]))

month_abb = numpy.array(
    [
        "Jan",
        "Feb",
        "Mar",
        "Apr",
        "May",
        "Jun",
        "Jul",
        "Aug",
        "Sep",
        "Oct",
        "Nov",
        "Dec",
    ]
)
month_name = numpy.array(
    [
        "January",
        "February",
        "March",
        "April",
        "May",
        "June",
        "July",
        "August",
        "September",
        "October",
        "November",
        "December",
    ]
)
